from __future__ import division

import numpy as np
import pandas as pd
import scipy.sparse as sps

from inferelator.distributed.inferelator_mp import MPControl
from inferelator.utils import Debug, InferelatorData, array_set_diag
from inferelator.utils import Validator as check

# Number of discrete bins for mutual information calculation
DEFAULT_NUM_BINS = 10

# DDOF for CLR
CLR_DDOF = 1

# Log type for MI calculations. np.log2 gives results in bits; np.log gives results in nats
DEFAULT_LOG_TYPE = np.log

# KVS keys for multiprocessing
SYNC_CLR_KEY = 'post_clr'


class MIDriver:

    @staticmethod
    def run(x, y, bins=DEFAULT_NUM_BINS, logtype=DEFAULT_LOG_TYPE, return_mi=True):
        return context_likelihood_mi(x, y, bins=bins, logtype=logtype, return_mi=return_mi)


def context_likelihood_mi(x, y, bins=DEFAULT_NUM_BINS, logtype=DEFAULT_LOG_TYPE, return_mi=True):
    """
    Wrapper to calculate the Context Likelihood of Relatedness and Mutual Information for two data sets that have
    common condition rows. The y argument will be used to calculate background MI for the x & y MI.
    As an implementation detail, y will be cast to a dense array if it is sparse.
    X can be sparse with no internal copy.

    This function handles unpacking and packing the InferelatorData.

    :param x: An N x G InferelatorData object
    :type x: InferelatorData [N x G]
    :param y: An N x K InferelatorData object
    :type y: InferelatorData [N x K]
    :param logtype: The logarithm function to use when calculating information. Defaults to natural log (np.log)
    :type logtype: np.log func
    :param bins: Number of bins for discretizing continuous variables
    :type bins: int
    :param return_mi: Boolean for returning a MI object. Defaults to True
    :type return_mi: bool
    :return clr, mi: CLR and MI InferelatorData objects. Returns (CLR, None) if return_mi is False.
    :rtype InferelatorData, InferelatorData:
    """

    assert check.argument_integer(bins, allow_none=True)
    assert min(x.shape) > 0
    assert min(y.shape) > 0
    assert check.indexes_align((x.sample_names, y.sample_names))

    # Create dense output matrix and copy the inputs
    mi_r = x.gene_names
    mi_c = y.gene_names

    # Build a [G x K] mutual information array
    mi = mutual_information(x.expression_data, y.expression_data, bins, logtype=logtype)
    array_set_diag(mi, 0., mi_r, mi_c)

    # Build a [K x K] mutual information array
    mi_bg = mutual_information(y.expression_data, y.expression_data, bins, logtype=logtype)
    array_set_diag(mi_bg, 0., mi_c, mi_c)

    # Calculate CLR
    clr = calc_mixed_clr(mi, mi_bg)

    MPControl.sync_processes(pref=SYNC_CLR_KEY)

    mi = pd.DataFrame(mi, index=mi_r, columns=mi_c)
    clr = pd.DataFrame(clr, index=mi_r, columns=mi_c)

    return clr, mi if return_mi else None


def mutual_information(x, y, bins, logtype=DEFAULT_LOG_TYPE):
    """
    Calculate the mutual information matrix between two data matrices, where the columns are equivalent conditions

    :param x: np.array (n x m1)
        The data from m1 variables across n conditions
    :param y: np.array (n x m2)
        The data from m2 variables across n conditions
    :param bins: int
        Number of bins to discretize continuous data into for the generation of a contingency table
    :param logtype: np.log func
        Which type of log function should be used (log2 results in MI bits, log results in MI nats, log10... is weird)

    :return mi: pd.DataFrame (m1 x m2)
        The mutual information between variables m1 and m2
    """

    # Discretize the input matrix y
    y = y.A if sps.isspmatrix(y) else y
    y = _make_array_discrete(y, bins, axis=0)

    # Build the MI matrix
    if MPControl.is_dask():
        from inferelator.distributed.dask_functions import build_mi_array_dask
        return build_mi_array_dask(x, y, bins, logtype=logtype)
    else:
        return build_mi_array(x, y, bins, logtype=logtype)


def build_mi_array(X, Y, bins, logtype=DEFAULT_LOG_TYPE, temp_dir=None):
    """
    Calculate MI into an array

    :param X: np.ndarray (n x m1)
        Discrete array of bins
    :param Y: np.ndarray (n x m2)
        Discrete array of bins
    :param bins: int
        The total number of bins that were used to make the arrays discrete
    :param logtype: np.log func
        Which log function to use (log2 gives bits, ln gives nats)
    :param temp_dir: path
        Path to write temp files for multiprocessing
    :return mi: np.ndarray (m1 x m2)
        Returns the mutual information array
    """

    m1, m2 = X.shape[1], Y.shape[1]

    # Define the function which calculates MI for each variable in X against every variable in Y
    def mi_make(i):
        level = 2 if i % 1000 == 0 else 3
        Debug.allprint("Mutual Information Calculation [{i} / {total}]".format(i=i, total=m1), level=level)

        discrete_X = _make_discrete(X[:, i].A.flatten() if sps.isspmatrix(X) else X[:, i].flatten(), bins)
        return [_calc_mi(_make_table(discrete_X, Y[:, j], bins), logtype=logtype) for j in range(m2)]

    # Send the MI build to the multiprocessing controller
    mi_list = MPControl.map(mi_make, range(m1), tmp_file_path=temp_dir)

    # Convert the list of lists to an array
    mi = np.array(mi_list)
    assert (m1, m2) == mi.shape, "Array {sh} produced [({m1}, {m2}) expected]".format(sh=mi.shape, m1=m1, m2=m2)

    return mi


def calc_mixed_clr(mi, mi_bg):
    """
    Calculate the context liklihood of relatedness from mutual information and the background mutual information

    :param mi: Mutual information array [m1 x m2]
    :type mi: np.ndarray
    :param mi_bg: Background mutual information array [m2 x m2]
    :type mi_bg: np.ndarray
    :return clr: Context liklihood of relateness array [m1 x m2]
    :rtype: np.ndarray
    """

    with np.errstate(invalid='ignore'):

        # Calculate the zscore for the dynamic CLR
        z_dyn = np.round(mi, 10)  # Rounding so that float precision differences don't turn into huge CLR differences
        z_dyn = np.subtract(z_dyn, np.mean(mi, axis=0))
        z_dyn = np.divide(z_dyn, np.std(mi, axis=0, ddof=CLR_DDOF))

        # Calculate the zscore for the static CLR
        z_stat = np.round(mi, 10)  # Rounding so that float precision differences don't turn into huge CLR differences
        z_stat = np.subtract(z_stat, np.mean(mi_bg, axis=0))
        z_stat = np.divide(z_stat, np.std(mi_bg, axis=0, ddof=CLR_DDOF))

        z_dyn[z_dyn < 0] = 0
        z_stat[z_stat < 0] = 0

    # Calculate CLR
    return np.sqrt(np.square(z_dyn) + np.square(z_stat))


def _make_array_discrete(array, num_bins, axis=0):
    """
    Applies _make_discrete to a 2d array
    """
    return np.apply_along_axis(_make_discrete, arr=array, axis=axis, num_bins=num_bins)


def _make_discrete(arr_vec, num_bins):
    """
    Takes a 1d array or vector and discretizes it into nonparametric bins
    :param arr_vec: np.ndarray
        1d array of continuous data
    :param num_bins: int
        Number of bins for data
    :return array: np.ndarray
        1d array of discrete data
    """

    # Sanity check
    assert check.argument_type(arr_vec, np.ndarray)
    assert len(arr_vec.shape) == 1 or arr_vec.shape[1] == 1

    # Get array min and max
    arr_min, arr_max = np.min(arr_vec), np.max(arr_vec)

    # Short circuit if the variance is 0
    if arr_min == arr_max:
        return np.zeros(shape=arr_vec.shape, dtype=np.dtype(int))

    # Continuous values to discrete bins [0, num_bins)
    # Write directly into a np.int16 array with the standard unsafe conversion
    return np.floor((arr_vec - arr_min) / (arr_max - arr_min + np.spacing(arr_max - arr_min)) * num_bins,
                    out=np.zeros(shape=arr_vec.shape, dtype=np.int16), casting='unsafe')


def _make_table(x, y, num_bins):
    """
    Takes two variable vectors which have been made into discrete integer bins and constructs a contingency table
    :param x: np.ndarray
        1d array of discrete data
    :param y: np.ndarray
        1d array of discrete data
    :param num_bins: int
        Number of bins for data
    :return ctable: np.ndarray (num_bins x num_bins)
        Contingency table of variables X and Y
    """

    assert len(x.shape) == 1
    assert len(y.shape) == 1

    # The only fast way to do this is by reindexing the table as an index array
    reindex = x * num_bins + y
    # Then piling everything up with bincount and reshaping it back into the table
    return np.bincount(reindex, minlength=num_bins ** 2).reshape(num_bins, num_bins).astype(np.dtype(float))


def _calc_mi(table, logtype=DEFAULT_LOG_TYPE):
    """
    Calculate Mutual Information from a contingency table of two variables

    :param table: np.ndarray (num_bins x num_bins)
        Contingency table
    :param logtype: np.log func
        Log function to use
    :return: float
        Mutual information between variable x & y
    """

    # Turn off runtime warnings (there is an explicit check for NaNs and INFs in-function)
    with np.errstate(divide='ignore', invalid='ignore'):
        m, n = table.shape
        assert n == m

        total = np.sum(table, axis=(0, 1))

        # (PxPy) [n x n]
        mi_val = np.dot((np.sum(table, axis=1) / total).reshape(-1, 1),
                        (np.sum(table, axis=0) / total).reshape(1, -1))

        # (Pxy) [n x n]
        table = np.divide(table, total)

        # (Pxy)/(PxPy) [n x n]
        mi_val = np.divide(table, mi_val)

        # log[(Pxy)/(PxPy)] [n x n]
        mi_val = logtype(mi_val)

        # Pxy(log[(Pxy)/(PxPy)]) [n x n]
        mi_val = np.multiply(table, mi_val)
        mi_val[np.isnan(mi_val)] = 0

        # Summation
        return np.sum(mi_val, axis=(0, 1))



"""
LocalController just runs everything in a single process
"""

import collections

from inferelator.distributed import AbstractController
from inferelator import utils
from inferelator.utils import Validator as check


class LocalController(AbstractController):

    _controller_name = "local"

    client = None
    is_master = True
    chunk = None

    @classmethod
    def connect(cls, *args, **kwargs):
        return True

    @classmethod
    def sync_processes(cls, *args, **kwargs):
        return True

    @classmethod
    def map(cls, func, *arg, **kwargs):
        """
        Map a function across iterable(s) and return a list of results

        :param func: function
            Mappable function
        :param args: iterable
            Iterator(s)
        """
        assert check.argument_callable(func)
        assert check.argument_list_type(arg, collections.Iterable)
        return list(map(func, *arg))

    @classmethod
    def set_processes(cls, process_count):
        """
        Set the number of dask workers to use
        :param process_count: int
        :return:
        """
        utils.Debug.vprint("Local does not support multiple cores", level=0)

    @classmethod
    def shutdown(cls):
        return True



from inferelator.distributed import AbstractController
from inferelator import utils
import warnings

DEFAULT_MP_ENGINE = "local"


class MPControl(AbstractController):
    """
    This is the multiprocessing controller. It is a pass-through for the method-specific multiprocessing implementations
    A multiprocessing implementation can be registered here and then used throughout the inferelator
    """

    _controller_name = "multiprocessing_registry"
    client = None

    # Relevant external state booleans
    is_master = False
    is_initialized = False

    @classmethod
    def name(cls):
        """
        This returns the _class_name which all subclasses should define
        """
        if cls.client is None:
            raise NameError("No client has been set")
        return cls.client.name()

    @classmethod
    def is_dask(cls):
        """
        This returns True if dask functions should be used
        """
        if cls.client is None:
            return False
        return cls.client.is_dask()

    @classmethod
    def set_multiprocess_engine(cls, engine, processes=None):
        """
        Register the multiprocessing engine to use

        Currently available are:

        dask-cluster
        dask-k8
        dask-local
        kvs
        multiprocessing
        local

        :param engine: A string to lookup the controller or a Controller object
        :type engine: str, Controller
        :param processes: Number of processes to use. Equivalent to calling `set_processes`
        :type processes: int
        """
        if cls.is_initialized:
            raise RuntimeError("Client is currently active. Run .shutdown() before changing engines.")

        if utils.is_string(engine):
            if engine == "dask-cluster":
                from inferelator.distributed.dask_cluster_controller import DaskHPCClusterController
                cls.client = DaskHPCClusterController
            elif engine == "dask-local":
                from inferelator.distributed.dask_local_controller import DaskController
                cls.client = DaskController
            elif engine == "dask-k8":
                from inferelator.distributed.dask_k8_controller import DaskK8Controller
                cls.client = DaskK8Controller
            elif engine == "kvs":
                warnings.warn("The KVS engine is deprecated. It has been replaced by Dask-based multiprocessing",
                              DeprecationWarning)
                from inferelator.distributed.kvs_controller import KVSController
                cls.client = KVSController
            elif engine == "multiprocessing":
                from inferelator.distributed.multiprocessing_controller import MultiprocessingController
                cls.client = MultiprocessingController
            elif engine == "local":
                from inferelator.distributed.local_controller import LocalController
                cls.client = LocalController
            else:
                raise ValueError("Engine {eng_str} unknown".format(eng_str=engine))
        elif issubclass(engine, AbstractController):
            cls.client = engine
        else:
            raise ValueError("Engine must be provided as a string for lookup or an implemented Controller class object")

        utils.Debug.vprint("Inferelator MPControl using engine {eng}".format(eng=cls.name()))

        if processes is not None:
            cls.set_processes(processes)

    @classmethod
    def connect(cls, *args, **kwargs):
        """
        Connect to the manager or scheduler or process pool or whatever using the `.connect()` implementation in the
        multiprocessing engine.
        """
        if cls.is_initialized:
            return True

        if cls.client is None:
            utils.Debug.vprint("Loading default engine {eng}".format(eng=DEFAULT_MP_ENGINE))
            cls.set_multiprocess_engine(DEFAULT_MP_ENGINE)

        connect_return = cls.client.connect(*args, **kwargs)

        # Set the process state
        cls.is_master = cls.client.is_master
        cls.is_initialized = True

        # Also tell Debug if we're the master process
        utils.Debug.is_master = cls.is_master

        return connect_return

    @classmethod
    def map(cls, func, *args, **kwargs):
        """
        Map using the `.map()` implementation in the multiprocessing engine
        """
        if not cls.is_initialized:
            raise RuntimeError("Connect before calling map()")
        return cls.client.map(func, *args, **kwargs)

    @classmethod
    def set_processes(cls, process_count):
        """
        Set worker process count
        """
        if cls.is_initialized:
            raise RuntimeError("Cannot set processes after the engine has started")
        return cls.client.set_processes(process_count)

    @classmethod
    def sync_processes(cls, *args, **kwargs):
        """
        Make sure processes are all at the same point by calling the `.sync_processes()` implementation in the
        multiprocessing engine

        This is necessary for KVS; other engines will just return True
        """
        if not cls.is_initialized:
            raise RuntimeError("Connect before calling sync_processes()")
        return cls.client.sync_processes(*args, **kwargs)

    @classmethod
    def shutdown(cls):
        """
        Gracefully shut down the multiprocessing engine by calling `.shutdown()`
        """

        if cls.is_initialized:
            client_off = cls.client.shutdown()
            cls.is_initialized = False
            cls.client = None
        else:
            client_off = True

        return client_off



from __future__ import print_function, unicode_literals, division

import pandas as pd
import pandas.api.types as pat
import os
import inspect

# Python 2/3 compatible string checking
try:
    basestring
except NameError:
    basestring = str


class Validator(object):
    """
    Validation module for function arguments. Each function here should return True or it should raise an exception
    """

    @staticmethod
    def argument_numeric(arg, low=None, high=None, allow_none=False, types=(int, float)):
        """
        Validate an input argument as being numeric (either an int or a float). Also check bounds if set.
        :param arg:
            Argument to validate
        :param low: numeric
            Lowest (inclusive) acceptable value of the argument; ignore this if it's None
        :param high: numeric
            Lowest (inclusive) acceptable value of the argument; ignore this if it's None
        :param allow_none: bool
            Allow arg to be None if true
        :return:
            Returns True if valid. Raises an exception otherwise
        """
        if allow_none and arg is None:
            return True

        if not isinstance(arg, types):
            raise ValueError("Argument must be numeric ({arg}, {typ} provided) ".format(arg=arg, typ=type(arg)))

        if low is not None and Validator.argument_numeric(low) and arg < low:
            raise ValueError("Argument must be at least {low}".format(low=low))
        if high is not None and Validator.argument_numeric(high) and arg > high:
            raise ValueError("Argument must be no more than {high}".format(high=high))

        return True

    @staticmethod
    def argument_integer(arg, low=None, high=None, allow_none=False):
        """
        Wrapper for argument_numeric which forces only integers
        """
        return Validator.argument_numeric(arg, low=low, high=high, allow_none=allow_none, types=int)

    @staticmethod
    def argument_enum(arg, enum_list, allow_none=False):
        """
        Validate an input argument as being present in an list of acceptable values
        :param arg:
            Argument to validate. If arg is a list or tuple, validate that each element is acceptable
        :param enum_list:
            A list or tuple (or anything that you can use 'in' with; like an index) of valid arguments
        :param allow_none: bool
            Allow arg to be None if true
        :return:
            Returns True if valid. Raises an exception otherwise
        """

        if allow_none and arg is None:
            return True

        if isinstance(arg, (list, tuple)):
            for a in arg:
                Validator.argument_enum(a, enum_list, allow_none=allow_none)
            return True
        elif arg not in enum_list:
            raise ValueError("Argument {arg} must be one of: {enum}".format(arg=arg, enum=",".join(enum_list)))
        else:
            return True

    @staticmethod
    def argument_path(arg, allow_none=False, create_if_needed=False, access=None):
        """
        Check to see if a path exists
        :param arg: str
            Path to a target
        :param allow_none: bool
            Allow arg to be None
        :param create_if_needed: bool
            Create a folder with os.makedirs if the path doesn't exist
        :param access:
            Mode parameter to check access
        :return:
            Returns True if valid. Raises an exception otherwise
        """
        if allow_none and arg is None:
            return True

        # If the path doesn't exist, create it or raise ValueError
        if not os.path.exists(arg) and create_if_needed:
            try:
                os.makedirs(arg)
            except OSError as err:
                raise ValueError("Path {arg} does not exist and cant be created:\n{err}".format(arg=arg, err=str(err)))
        elif not os.path.exists(arg):
            raise ValueError("Argument {arg} must be an existing path".format(arg=arg))

        # If access is set, check and see if the permissions are OK and raise ValueError if not
        if access is not None:
            if os.access(arg, access):
                return True
            else:
                raise ValueError("Path {arg} does not have permission {per}".format(arg=arg, per=access))
        else:
            return True

    @staticmethod
    def argument_subpath(arg, is_subpath_of, allow_none=False):
        """
        Check and see if an argument is a subpath of a path
        :param arg: str
            Path
        :param is_subpath_of: str
            Path for comparison
        :param allow_none: bool
            Allow arg to be None
        :return:
        """

        if allow_none and arg is None:
            return True
        elif arg is None or is_subpath_of is None:
            raise ValueError("Path argument {a} cannot be compared to path {b}".format(a=arg, b=is_subpath_of))

        arg = os.path.abspath(os.path.expanduser(arg))
        is_subpath_of = os.path.abspath(os.path.expanduser(is_subpath_of))

        if arg == is_subpath_of:
            return True
        elif is_subpath_of == os.path.abspath(os.sep):
            return True
        elif arg.startswith(is_subpath_of + os.sep):
            return True
        else:
            raise ValueError("Path {a} is not a subpath of path {b}".format(a=arg, b=is_subpath_of))

    @staticmethod
    def argument_type(arg, arg_type, allow_none=False):
        if allow_none and arg is None:
            return True

        if isinstance(arg, arg_type):
            return True
        else:
            raise ValueError("Argument {arg} must be of type {typ}".format(arg=arg, typ=arg_type))

    @staticmethod
    def argument_string(arg, allow_none=False):
        return Validator.argument_type(arg, basestring, allow_none=allow_none)

    @staticmethod
    def argument_list_type(arg, arg_type, allow_none=False):
        if allow_none and arg is None:
            return True
        for a in arg:
            Validator.argument_type(a, arg_type, allow_none=allow_none)
        return True

    @staticmethod
    def argument_callable(arg, allow_none=False):
        if allow_none and arg is None:
            return True
        elif callable(arg):
            return True
        else:
            raise ValueError("Argument {arg} must be callable".format(arg=arg))

    @staticmethod
    def dataframes_align(frame_iterable, allow_none=False, check_order=True):

        is_none = [f is None for f in frame_iterable]
        if any(is_none) and allow_none:
            # If None is an allowed value, remove the Nones and check the remaining dataframes
            new_frame_iterable = []
            for frame in frame_iterable:
                if frame is not None:
                    new_frame_iterable.append(frame)

            # If there are any non-None dataframes, check them for alignment. Otherwise return True
            if len(new_frame_iterable) > 0:
                frame_iterable = new_frame_iterable
            else:
                return True
        elif any(is_none):
            # If None isn't allowed, throw an error
            raise ValueError("None values are present in dataframe list")

        try:
            Validator.indexes_align([f.index for f in frame_iterable], allow_none=allow_none, check_order=check_order)
        except ValueError as ve:
            raise ValueError("Dataframes are not aligned on indexes: {err}".format(err=str(ve)))

        try:
            Validator.indexes_align([f.columns for f in frame_iterable], allow_none=allow_none, check_order=check_order)
        except ValueError as ve:
            raise ValueError("Dataframes are not aligned on columns: {err}".format(err=str(ve)))

        return True

    @staticmethod
    def dataframe_is_numeric(frame, allow_none=False):
        if allow_none and frame is None:
            return True

        non_numeric = pd.Index([not pat.is_numeric_dtype(x) for x in frame.dtypes])

        if non_numeric.any():
            bad_features = "\t".join(map(str, frame.columns[non_numeric].tolist()))
            raise ValueError("Dataframe has non-numeric features: {f}".format(f=bad_features))
        else:
            return True

    @staticmethod
    def dataframe_is_finite(frame, allow_none=False, check_index=True):
        if allow_none and frame is None:
            return True

        with pd.option_context('mode.use_inf_as_na', True):
            non_finites = frame.apply(lambda x: pd.isnull(x).sum()) > 0
            if non_finites.any():
                bad_features = "\t".join(map(str, frame.columns[non_finites].tolist()))
                raise ValueError("Dataframe has non-finite features: {f}".format(f=bad_features))
            elif check_index and pd.isnull(frame.index).any():
                raise ValueError("NaN values are present in frame index")
            elif check_index and pd.isnull(frame.columns).any():
                raise ValueError("NaN values are present in frame column")
            else:
                return True

    @staticmethod
    def indexes_align(index_iterable, allow_none=False, check_order=True):
        is_none = [f is None for f in index_iterable]
        if any(is_none) and allow_none:
            # If None is an allowed value, remove the Nones and check the remaining dataframes
            new_index_iterable = []
            for index in index_iterable:
                if index is not None:
                    new_index_iterable.append(index)

            # If there are any non-None dataframes, check them for alignment. Otherwise return True
            if len(new_index_iterable) > 0:
                index_iterable = new_index_iterable
            else:
                return True
        elif any(is_none):
            # If None isn't allowed, throw an error
            raise ValueError("None values are present in dataframe list")

        order_flag = False
        zindex = index_iterable[0]
        for ind in index_iterable:
            if len(zindex.difference(ind)) > 0:
                raise ValueError("Indexes have mismatching labels: "+"\t".join(map(str, zindex.difference(ind))))
            elif check_order and any(zindex != ind):
                order_flag = True

        if order_flag:
            raise ValueError("Indexes have matching labels but mismatching order")

        return True

    @staticmethod
    def index_values_unique(index, allow_none=False):
        """
        Check and make sure a pandas index doesn't have duplicate entries
        :param index:
        :param allow_none:
        :return:
        """
        if allow_none and index is None:
            return True
        elif index is None:
            raise ValueError("None is not an acceptable argument")
        elif index.duplicated().sum() > 0:
            dupes = index[index.duplicated()].tolist()
            raise ValueError("Duplicate value(s) present in index: {dupes}".format(dupes=" ".join(dupes)))
        else:
            return True

    @staticmethod
    def arguments_not_none(args, num_none=None):
        """
        :param args:
            Tuple of arguments to check
        :param num_none: int
            The number of arguments which should not be None (so 1 means exactly 1 argument should be not None)
            If None, all arguments should not be None
        """
        n_not_none = 0
        for ar in args:
            n_not_none += 0 if ar is None else 1

        if num_none is None and n_not_none != len(args):
            raise ValueError("One of these arguments is None; None is not an acceptable argument")
        elif num_none is not None and n_not_none != num_none:
            raise ValueError("{num} arguments are not None; only {nnum} are allowed".format(num=n_not_none,
                                                                                            nnum=num_none))
        return True

    @staticmethod
    def argument_is_subclass(arg, subclass, allow_none=False):
        """

        :param arg:
            Argument to check
        :param subclass:
            Class that the argument must be a subclass of
        :param allow_none:
        """
        if allow_none and arg is None:
            return True
        elif arg is None:
            raise ValueError("None is not an acceptable argument")

        if not inspect.isclass(arg) and inspect.isclass(type(arg)):
            arg = type(arg)

        if not inspect.isclass(subclass):
            raise ValueError("Subclass to test argument is itself not a class")
        elif not issubclass(arg, subclass):
            raise ValueError("Argument is not a subclass of {sc}".format(sc=str(subclass)))
        else:
            return True


def is_string(arg):
    """
    Check if a argument is a string in a python 2/3 compatible way
    :param arg:
    :return:
    """
    return isinstance(arg, basestring)



from inferelator.postprocessing.column_names import *
from inferelator.postprocessing.model_metrics import MetricHandler, RankSummingMetric, CombinedMetric
from inferelator.postprocessing.inferelator_results import InferelatorResults
from inferelator.postprocessing.results_processor import ResultsProcessor
from inferelator.postprocessing.results_processor_mtl import ResultsProcessorMultiTask



# This is the required package init. Everything in this package must implement the abstract class AbstractController

from abc import abstractmethod


class AbstractController:
    # The object which handles the multiprocessing
    client = None

    # Boolean to identify master processes where needed
    is_master = False

    # The chunk sizes for calls to map
    chunk = 25

    # The name of this controller
    _controller_name = None
    _controller_dask = False

    @classmethod
    def name(cls):
        """
        This returns the _class_name which all subclasses should define
        """
        if cls._controller_name is None:
            raise NameError("Controller name has not been defined")
        return cls._controller_name

    @classmethod
    def is_dask(cls):
        return cls._controller_dask

    @classmethod
    @abstractmethod
    def connect(cls, *args, **kwargs):
        """
        This establishes or creates a multiprocessing state that allows `map` or other functionality to be used
        in parallel
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def map(cls, *args, **kwargs):
        """
        This implements a map function that will execute in parallel. If this is not implemented, the standard
        workflow will fail and provision must be made elsewhere to define the multiprocessing functionality
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def set_processes(cls, process_count):
        """
        Set the number of processes to use. Return warnings if the number of processes cannot be set
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def sync_processes(cls, *args, **kwargs):
        """
        This synchronizes multiple processes. Multiprocessing methods which have a defined hierarchy and no risk of
        race conditions may simply return True
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def shutdown(cls):
        """
        Clean shutdown of the multiprocessing state
        """
        raise NotImplementedError



import numpy as np
import math
from scipy import sparse as _sparse

from inferelator.utils import Debug
from inferelator.distributed.inferelator_mp import MPControl


def make_data_noisy(data, random_seed=42):
    """
    Generate a new data object of random data which matches the provided data

    :param data: Raw read data
    :type data: InferelatorData
    :param random_seed: Random seed for data generation
    :type random_seed: int
    :return: Simulated data
    :rtype: InferelatorData
    """

    # Calculate probability vector for gene expression
    # Discrete sampling for count data

    sample_counts = data.sample_counts

    if data._is_integer:

        Debug.vprint("Simulating integer count data for {n} samples".format(n=data.num_obs), level=0)

        # Data is centered already
        if np.any(sample_counts <= 0.):
            p_vec = np.ones(data.num_genes, dtype=float)

        # Normalize to mean counts per sample and sum counts per gene by matrix multiplication
        else:
            p_vec = (np.mean(sample_counts) / sample_counts).reshape(1, -1) @ data.expression_data

        # Flatten and convert counts to a probability vector
        p_vec = p_vec.flatten()
        p_vec = p_vec / p_vec.sum()

        data.expression_data = _sim_ints(p_vec, sample_counts, sparse=data.is_sparse, random_seed=random_seed)

    else:

        # Data is centered already
        if np.any(sample_counts <= 0.):
            p_vec = np.zeros(data.num_genes, dtype=float)

        # Normalize to mean total measured values per sample and sum counts per gene by matrix multiplication
        else:
            p_vec = (np.mean(sample_counts) / sample_counts).reshape(1, -1) @ data.expression_data
            p_vec /= data.num_obs

        Debug.vprint("Simulating float data for {n} samples".format(n=data.num_obs), level=0)
        data.expression_data = _sim_float(p_vec.flatten(), data.gene_stdev, data.num_obs, random_seed=random_seed)


def _sim_ints(prob_dist, n_per_row, sparse=False, random_seed=42):

    if not np.isclose(np.sum(prob_dist), 1.):
        raise ValueError("Probability distribution does not sum to 1")

    ncols = len(prob_dist)

    def _sim_rows(n_vec, seed):
        row_data = np.zeros((len(n_vec), ncols), dtype=np.int32)

        rng = np.random.default_rng(seed=seed)
        col_ids = np.arange(ncols)

        for i, n in enumerate(n_vec):
            row_data[i, :] = np.bincount(rng.choice(col_ids, size=n, p=prob_dist), minlength=ncols)

        return _sparse.csr_matrix(row_data) if sparse else row_data

    ss = np.random.SeedSequence(random_seed)
    sim_data = MPControl.map(_sim_rows, _row_gen(n_per_row), _ss_gen(ss))

    return _sparse.vstack(sim_data) if sparse else np.vstack(sim_data)


def _sim_float(gene_centers, gene_sds, nrows, random_seed=42):

    ncols = len(gene_centers)
    assert ncols == len(gene_sds)

    def _sim_cols(cents, sds, seed):
        rng = np.random.default_rng(seed=seed)
        return rng.normal(loc=cents, scale=sds, size=(nrows, len(cents)))

    ss = np.random.SeedSequence(random_seed)

    return np.hstack(MPControl.map(_sim_cols, _col_gen(gene_centers), _col_gen(gene_sds), _ss_gen(ss)))


def _row_gen(n_vec, chunksize=2000):
    _chunks = math.ceil(len(n_vec) / chunksize)
    for i in range(_chunks):
        yield n_vec[i * chunksize: min(len(n_vec), (i + 1) * chunksize)]


def _col_gen(vals, chunksize=200):
    _chunks = math.ceil(len(vals) / chunksize)
    for i in range(_chunks):
        _start, _stop = i * chunksize, min(len(vals), (i + 1) * chunksize)
        yield vals[_start: _stop]


def _ss_gen(ss):
    while True:
        yield ss.generate_state(1)[0]




import pandas as pd
import numpy as np

from inferelator import utils
from inferelator.regression import bayes_stats
from inferelator.regression import base_regression
from inferelator.regression import mi
from inferelator.distributed.inferelator_mp import MPControl

# Default number of predictors to include in the model
DEFAULT_nS = 10

# Default weight for priors & Non-priors
# If prior_weight is the same as no_prior_weight:
#   Priors will be included in the pp matrix before the number of predictors is reduced to nS
#   They won't get special treatment in the model though
DEFAULT_prior_weight = 1
DEFAULT_no_prior_weight = 1

# Throw away the priors which have a CLR that is 0 before the number of predictors is reduced by BIC
DEFAULT_filter_priors_for_clr = False


class BBSR(base_regression.BaseRegression):
    # Bayseian correlation measurements

    # Priors Data
    prior_mat = None  # [G x K] # numeric
    filter_priors_for_clr = DEFAULT_filter_priors_for_clr  # bool

    # Weights for Predictors (weights_mat is set with _calc_weight_matrix)
    weights_mat = None  # [G x K] numeric
    prior_weight = DEFAULT_prior_weight  # numeric
    no_prior_weight = DEFAULT_no_prior_weight  # numeric

    # Predictors to include in modeling (pp is set with _build_pp_matrix)
    pp = None  # [G x K] bool
    nS = DEFAULT_nS  # int

    ols_only = False

    def __init__(self, X, Y, clr_mat, prior_mat, nS=DEFAULT_nS, prior_weight=DEFAULT_prior_weight,
                 no_prior_weight=DEFAULT_no_prior_weight, ordinary_least_squares=False):
        """
        Create a Regression object for Bayes Best Subset Regression

        :param X: Expression or Activity data [N x K]
        :type X: InferelatorData
        :param Y: Response expression data [N x G]
        :type Y: InferelatorData
        :param clr_mat: Calculated CLR between features of X & Y [G x K]
        :type clr_mat: pd.DataFrame
        :param prior_mat: Prior data between features of X & Y [G x K]
        :type prior_mat: pd.DataFrame

        :param nS: int
            Number of predictors to retain
        :param prior_weight: int
            Weight of a predictor which does have a prior
        :param no_prior_weight: int
            Weight of a predictor which doesn't have a prior
        """

        super(BBSR, self).__init__(X, Y)

        self.nS = nS
        self.ols_only = ordinary_least_squares

        # Calculate the weight matrix
        self.prior_weight = prior_weight
        self.no_prior_weight = no_prior_weight
        weights_mat = self._calculate_weight_matrix(prior_mat, p_weight=prior_weight, no_p_weight=no_prior_weight)
        utils.Debug.vprint("Weight matrix {} construction complete".format(weights_mat.shape))

        # Rebuild weights, priors, and the CLR matrix for the features that are in this bootstrap
        self.weights_mat = weights_mat.loc[self.genes, self.tfs]
        self.prior_mat = prior_mat.loc[self.genes, self.tfs]
        self.clr_mat = clr_mat.loc[self.genes, self.tfs]

        # Build a boolean matrix indicating which tfs should be used as predictors for regression for each gene
        self.pp = self._build_pp_matrix()

    def regress(self):
        """
        Execute BBSR

        :return: pd.DataFrame [G x K], pd.DataFrame [G x K]
            Returns the regression betas and beta error reductions for all threads if this is the master thread (rank 0)
            Returns None, None if it's a subordinate thread
        """

        if MPControl.is_dask():
            from inferelator.distributed.dask_functions import bbsr_regress_dask
            return bbsr_regress_dask(self.X, self.Y, self.pp, self.weights_mat, self.G, self.genes, self.nS)

        def regression_maker(j):
            level = 0 if j % 100 == 0 else 2
            utils.Debug.allprint(base_regression.PROGRESS_STR.format(gn=self.genes[j], i=j, total=self.G),
                                 level=level)

            data = bayes_stats.bbsr(self.X.values,
                                    utils.scale_vector(self.Y.get_gene_data(j, force_dense=True).flatten()),
                                    self.pp.iloc[j, :].values.flatten(),
                                    self.weights_mat.iloc[j, :].values.flatten(),
                                    self.nS,
                                    ordinary_least_squares=self.ols_only)
            data['ind'] = j
            return data

        return MPControl.map(regression_maker, range(self.G), tell_children=False)

    def _build_pp_matrix(self):
        """
        From priors and context likelihood of relatedness, determine which predictors should be included in the model
        :return pp: pd.DataFrame [G x K]
            Boolean matrix indicating which predictor variables should be included in BBSR for each response variable
        """

        # Create a predictor boolean array from priors
        pp = np.logical_or(self.prior_mat != 0, self.weights_mat != self.no_prior_weight)

        pp_idx = pp.index
        pp_col = pp.columns

        if self.filter_priors_for_clr:
            # Set priors which have a CLR of 0 to FALSE
            pp = np.logical_and(pp, self.clr_mat != 0).values
        else:
            pp = pp.values

        # Mark the nS predictors with the highest CLR true (Do not include anything with a CLR of 0)
        mask = np.logical_or(self.clr_mat == 0, ~np.isfinite(self.clr_mat)).values
        masked_clr = np.ma.array(self.clr_mat.values, mask=mask)
        for i in range(self.G):
            n_to_keep = min(self.nS, self.K, mask.shape[1] - np.sum(mask[i, :]))
            if n_to_keep == 0:
                continue
            clrs = np.ma.argsort(masked_clr[i, :], endwith=False)[-1 * n_to_keep:]
            pp[i, clrs] = True

        # Rebuild into a DataFrame and set autoregulation to 0
        pp = pd.DataFrame(pp, index=pp_idx, columns=pp_col, dtype=np.dtype(bool))
        pp = utils.df_set_diag(pp, False)

        return pp

    @staticmethod
    def _calculate_weight_matrix(p_matrix, no_p_weight=DEFAULT_no_prior_weight,
                                 p_weight=DEFAULT_prior_weight):
        """
        Create a weights matrix. Everywhere p_matrix is not set to 0, the weights matrix will have p_weight. Everywhere
        p_matrix is set to 0, the weights matrix will have no_p_weight
        :param p_matrix: pd.DataFrame [G x K]
        :param no_p_weight: int
            Weight of something which doesn't have a prior
        :param p_weight: int
            Weight of something which does have a prior
        :return weights_mat: pd.DataFrame [G x K]
        """
        weights_mat = p_matrix * 0 + no_p_weight
        return weights_mat.mask(p_matrix != 0, other=p_weight)


class BBSRRegressionWorkflowMixin(base_regression._RegressionWorkflowMixin):
    """
    Bayesian Best Subset Regression (BBSR)

    https://doi.org/10.15252/msb.20156236
    """

    mi_driver = mi.MIDriver
    mi_sync_path = None

    prior_weight = DEFAULT_prior_weight
    no_prior_weight = DEFAULT_no_prior_weight
    bsr_feature_num = DEFAULT_nS
    clr_only = False
    ols_only = False

    def set_regression_parameters(self, prior_weight=None, no_prior_weight=None, bsr_feature_num=None, clr_only=False,
                                  ordinary_least_squares_only=None):
        """
        Set regression parameters for BBSR

        :param prior_weight: Weight for edges that are present in the prior network. Defaults to 1.
        :type prior_weight: float
        :param no_prior_weight: Weight for edges that are not present in the prior network. Defaults to 1.
        :type no_prior_weight: float
        :param bsr_feature_num: The number of features to include in best subset regression. Defaults to 10.
        :type bsr_feature_num: int
        :param clr_only: Only use Context Likelihood of Relatedness to select features for BSR, not prior edges.
            Defaults to False.
        :type clr_only: bool
        :param ordinary_least_squares_only: Use OLS instead of Bayesian regression, for testing. Defaults to False.
        :type ordinary_least_squares_only: bool
        """

        self._set_with_warning("prior_weight", prior_weight)
        self._set_with_warning("no_prior_weight", no_prior_weight)
        self._set_with_warning("bsr_feature_num", bsr_feature_num)
        self._set_without_warning("clr_only", clr_only)
        self._set_without_warning("ols_only", ordinary_least_squares_only)

    def run_bootstrap(self, bootstrap):
        X = self.design.get_bootstrap(bootstrap)
        Y = self.response.get_bootstrap(bootstrap)

        utils.Debug.vprint('Calculating MI, Background MI, and CLR Matrix', level=0)
        clr_matrix, _ = self.mi_driver().run(Y, X, return_mi=False)
        utils.Debug.vprint('Calculating betas using BBSR', level=0)

        # Create a mock prior with no information if clr_only is set
        if self.clr_only:
            priors = pd.DataFrame(0, index=self.priors_data.index, columns=self.priors_data.columns)
        else:
            priors = self.priors_data

        return BBSR(X, Y, clr_matrix, priors, prior_weight=self.prior_weight,
                    no_prior_weight=self.no_prior_weight, nS=self.bsr_feature_num,
                    ordinary_least_squares=self.ols_only).run()



import numpy as np

from inferelator.postprocessing.model_performance import RankSummingMetric
from inferelator.postprocessing import (TARGET_COLUMN, REGULATOR_COLUMN, CONFIDENCE_COLUMN, GOLD_STANDARD_COLUMN,
                                        MCC_COLUMN, TP, FP, TN, FN)

import matplotlib
matplotlib.use('pdf')
import matplotlib.pyplot as plt


class RankSummaryMCC(RankSummingMetric):
    """
    This class extends RankSumming and calculates Matthews correlation coefficient
    """

    name = "MCC"
    curve_file_name = "mccVSconf_curve.pdf"

    @property
    def mcc(self):
        return self.maxmcc

    @property
    def optconfmcc(self):
        return RankSummaryMCC.calculate_opt_conf_mcc(self.filtered_data)

    @property
    def maxmcc(self):
        return self.calculate_opt_mcc(self.filtered_data)

    @property
    def nnzmmc(self):
        return self.calculate_nnz_mcc(self.filtered_data, self.optconfmcc)

    # Plotter function

    def __init__(self, rankable_data, gold_standard, filter_method='keep_all_gold_standard'):

        super(RankSummaryMCC, self).__init__(rankable_data, gold_standard, filter_method=filter_method)

        # Calculate the precision and recall and store them with confidence data
        self.filtered_data = self.calculate_mcc(self.filtered_data.copy())

        # Join the filtered MCC onto the full confidences
        join_data = self.filtered_data.loc[:, [TARGET_COLUMN, REGULATOR_COLUMN, MCC_COLUMN]]
        join_data = join_data.set_index([TARGET_COLUMN, REGULATOR_COLUMN])
        self.confidence_data = self.confidence_data.join(join_data, on=[TARGET_COLUMN, REGULATOR_COLUMN])

    def score(self):

        return self.name, self.maxmcc

    def curve_dataframe(self):

        return self.filtered_data[[CONFIDENCE_COLUMN, MCC_COLUMN]]

    def output_curve(self, ax=None, figsize=(6, 4)):

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize, constrained_layout=True)
            ax = fig.add_subplot(1, 1, 1)

        # Extract the recall and precision data
        curve = self.curve_dataframe()
        self.plot_mcc_conf(curve[MCC_COLUMN].values, curve[CONFIDENCE_COLUMN].values, self.maxmcc, self.optconfmcc, ax,
                           num_edges=(self.confidence_data[CONFIDENCE_COLUMN] >= self.optconfmcc).sum())

        return ax

    @staticmethod
    def plot_mcc_conf(mcc, conf, optmcc, optconf, ax, num_edges=None):

        num_edges = np.sum(conf >= optconf) if num_edges is None else num_edges

        # Generate a plot
        ax.plot(conf, mcc)
        ax.set_xlabel('Confidence')
        ax.set_xlim(1, 0)
        ax.set_ylim(0, 1)
        ax.set_ylabel('MCC')
        ax.vlines(float(optconf), 0, 1, transform=ax.get_xaxis_transform(), colors='r', linestyles='dashed')

        _msg = "max MCC = {optmcc:.4f}\noptimal conf = {optconf:.4f}\nnum_edges = {n}".format(optmcc=optmcc,
                                                                                              optconf=optconf,
                                                                                              n=num_edges)
        ax.annotate(_msg, xy=(0.4, 0.075), xycoords='axes fraction')

        return ax

    @staticmethod
    def calculate_opt_mcc(data):

        return np.nanmax(data[MCC_COLUMN])

    @staticmethod
    def calculate_opt_conf_mcc(data):

        return data.loc[data[MCC_COLUMN] >= np.max(data[MCC_COLUMN]), CONFIDENCE_COLUMN].min()

    @staticmethod
    def calculate_nnz_mcc(data, conf):

        return (data[CONFIDENCE_COLUMN] >= conf).sum()

    @staticmethod
    def calculate_mcc(data):

        df = RankSummingMetric.compute_confusion_matrix(data)
        data[MCC_COLUMN] = RankSummaryMCC.confusion_to_mcc(df[TP], df[TN], df[FP], df[FN])
        return data

    @staticmethod
    def confusion_to_mcc(tp, tn, fp, fn):
        return (tp * tn - fp * fn) / (np.sqrt(tp + fp) * np.sqrt(tp + fn) * np.sqrt(tn + fp) * np.sqrt(tn + fn))



import pandas as pd
import os

from inferelator.utils import Validator as check
_SERIALIZE_ATTRS = ["network", "betas", "betas_stack", "betas_sign", "combined_confidences", "tasks"]


class InferelatorResults(object):
    """
    For network analysis, the results produced in the output_dir are sufficient.
    Model development and comparisons may require to values that are not written to files.
    An InferelatorResults object is returned by the ``workflow.run()`` methods
    (A list of InferelatorResults objects is returned by the ``CrossValidationManager.run()`` method).

    This object allows access to most of the internal values created by the inferelator.
    """

    #: Results name, usually set to task name.
    #: Defaults to None.
    name = None

    #: Network dataframe, usually written to network.tsv.gz
    network = None

    #: Fit model coefficients for each model bootstrap.
    #: This is a list of dataframes which are Genes x TFs.
    betas = None

    #: Count of non-zero betas.
    #: This is a dataframe which is Genes x TFs
    betas_stack = None

    #: The aggregate sign of non-zero betas.
    #: This is a dataframe which is Genes x TFs
    betas_sign = None

    #: Confidence scores for tf-gene network edges.
    #: This is a dataframe which is Genes x TFs
    combined_confidences = None

    #: Task result objects if there were multiple tasks. None if there were not.
    #: This is a dict, keyed by task ID
    tasks = None

    # File names
    network_file_name = "network.tsv.gz"
    confidence_file_name = "combined_confidences.tsv.gz"
    threshold_file_name = None
    curve_file_name = "combined_metrics.pdf"
    curve_data_file_name = None

    # Performance metrics
    metric = None
    curve = None
    score = None

    # Performance metrics - all scores
    all_scores = None
    all_names = None

    def __init__(self, network_data, betas_stack, combined_confidences, metric_object, betas_sign=None, betas=None):
        self.network = network_data
        self.betas_stack = betas_stack
        self.combined_confidences = combined_confidences
        self.metric = metric_object
        self.curve = metric_object.curve_dataframe()
        _, self.score = metric_object.score()
        self.all_names = metric_object.all_names()
        self.all_scores = metric_object.all_scores()
        self.betas_sign = betas_sign
        self.betas = betas

    def new_metric(self, metric_object, curve_file_name=None, curve_data_file_name=None):
        """
        Generate a new result object with a new metric
        :param metric_object:
        :param curve_file_name:
        :param curve_data_file_name:
        :return:
        """
        new_result = InferelatorResults(self.network, self.betas_stack, self.combined_confidences, metric_object,
                                        betas_sign=self.betas_sign, betas=self.betas)

        new_result.curve_data_file_name = curve_data_file_name
        new_result.curve_file_name = curve_file_name

        return new_result

    def plot_other_metric(self, metric_object, output_dir, curve_file_name=None, curve_data_file_name=None):
        """
        Write just the curve files for another provided metric.

        :param metric_object:
        :param output_dir:
        :param curve_file_name:
        :param curve_data_file_name:
        :return:
        """
        nm = self.new_metric(metric_object, curve_file_name=curve_file_name, curve_data_file_name=curve_data_file_name)
        nm.metric.output_curve_pdf(output_dir, curve_file_name) if curve_file_name is not None else None
        self.write_to_tsv(nm.curve, output_dir, curve_data_file_name) if curve_data_file_name is not None else None

    def write_result_files(self, output_dir):
        """
        Write all of the output files. Any individual file output can be overridden by setting the file name to None.
        All files can be suppressed by calling output_dir as None

        :param output_dir: Path to a directory where output files should be created. If None, no files will be made
        :type output_dir: str, None
        """

        # Validate that the output path exists (create it if necessary)
        check.argument_path(output_dir, allow_none=True, create_if_needed=True)

        # Write TSV files
        self.write_to_tsv(self.network, output_dir, self.network_file_name, index=False)
        self.write_to_tsv(self.combined_confidences, output_dir, self.confidence_file_name)
        self.write_to_tsv(self.betas_stack, output_dir, self.threshold_file_name, index=False)
        self.write_to_tsv(self.curve, output_dir, self.curve_data_file_name, index=False)

        # Write Metric Curve PDF
        self.metric.output_curve_pdf(output_dir, self.curve_file_name) if self.curve_file_name is not None else None

    def clear_output_file_names(self):
        """
        Reset the output file names (nothing will be output if this is called, unless new file names are set)
        """

        self.network_file_name, self.confidence_file_name, self.threshold_file_name = None, None, None
        self.curve_file_name, self.curve_data_file_name = None, None

    def save(self, output_dir, output_file_name):
        """
        Save the InferelatorResults to an HDF5 file

        :param output_dir:
        :param output_file_name:
        """

        if output_dir is None or output_file_name is None:
            return None

        with pd.HDFStore(os.path.join(output_dir, output_file_name)) as hdf5_store:

            # Save object dataframes
            for k in _SERIALIZE_ATTRS:
                if getattr(self, k) is not None:
                    hdf5_store.put(k, getattr(self, k))

            # If tasks exist, save task dataframes
            if self.tasks is not None:
                tasks = pd.Series(self.tasks.keys())
                hdf5_store.put("tasks", tasks)

                for t in tasks:
                    for k in _SERIALIZE_ATTRS:
                        if getattr(self.tasks[t], k) is not None:
                            hdf5_store.put(str(t) + "_" + str(k), getattr(self.tasks[t], k))

    @staticmethod
    def write_to_tsv(data_frame, output_dir, output_file_name, index=False, float_format='%.6f'):
        """
        Save a DataFrame to a TSV file

        :param data_frame: pd.DataFrame
            Data to write
        :param output_dir: str
            The path to the output file. If None, don't save anything
        :param output_file_name: str
            The output file name. If None, don't save anything
        :param index: bool
            Include the index in the output file
        :param float_format: str
            Reformat floats. Set to None to disable.
        """

        assert check.argument_type(data_frame, pd.DataFrame, allow_none=True)
        assert check.argument_path(output_dir, allow_none=True)
        assert check.argument_type(output_file_name, str, allow_none=True)

        # Write output
        if output_dir is not None and output_file_name is not None and data_frame is not None:
            data_frame.to_csv(os.path.join(output_dir, output_file_name), sep="\t", index=index, header=True,
                              float_format=float_format)



from inferelator.distributed.inferelator_mp import MPControl
from inferelator.regression import base_regression
from inferelator import utils
import copy

import numpy as np
import scipy.sparse as sps
from dask import distributed

"""
This package contains the dask-specific multiprocessing functions (these are used in place of map calls to allow the
more advanced memory and task tools of dask to be used)
"""

DASK_SCATTER_TIMEOUT = 120


def amusr_regress_dask(X, Y, priors, prior_weight, n_tasks, genes, tfs, G, remove_autoregulation=True,
                       lambda_Bs=None, lambda_Ss=None, Cs=None, Ss=None, regression_function=None):
    """
    Execute multitask (AMUSR)

    :return: list
        Returns a list of regression results that the amusr_regression pileup_data can process
    """

    assert MPControl.is_dask()

    from inferelator.regression.amusr_regression import format_prior, run_regression_EBIC
    DaskController = MPControl.client

    # Allows injecting a regression function for testing
    regression_function = run_regression_EBIC if regression_function is None else regression_function

    # Gets genes, n_tasks, prior_weight, and remove_autoregulation from regress_dask()
    # Other arguments are passed in
    def regression_maker(j, x_df, y_list, prior, tf):
        level = 0 if j % 100 == 0 else 2
        utils.Debug.allprint(base_regression.PROGRESS_STR.format(gn=genes[j], i=j, total=G),
                             level=level)

        gene = genes[j]
        x, y, tasks = [], [], []

        if remove_autoregulation:
            tf = [t for t in tf if t != gene]
        else:
            pass

        for k, y_data in y_list:
            x.append(x_df[k].get_gene_data(tf))  # list([N, K])
            y.append(y_data)
            tasks.append(k)  # [T,]

        prior = format_prior(prior, gene, tasks, prior_weight)
        return j, regression_function(x, y, tf, tasks, gene, prior,
                                      lambda_Bs=lambda_Bs, lambda_Ss=lambda_Ss, Cs=Cs, Ss=Ss)

    def response_maker(y_df, i):
        y = []
        gene = genes[i]
        for k in range(n_tasks):
            if gene in y_df[k].gene_names:
                y.append((k, y_df[k].get_gene_data(gene, force_dense=True).reshape(-1, 1)))
        return y

    # Scatter common data to workers
    [scatter_x] = DaskController.client.scatter([X], broadcast=True, hash=False)
    [scatter_priors] = DaskController.client.scatter([priors], broadcast=True, hash=False)

    # Wait for scattering to finish before creating futures
    distributed.wait(scatter_x, timeout=DASK_SCATTER_TIMEOUT)
    distributed.wait(scatter_priors, timeout=DASK_SCATTER_TIMEOUT)

    future_list = [DaskController.client.submit(regression_maker, i, scatter_x, response_maker(Y, i), scatter_priors,
                                                tfs)
                   for i in range(G)]

    # Collect results as they finish instead of waiting for all workers to be done
    result_list = process_futures_into_list(future_list)

    DaskController.client.cancel(scatter_x)
    DaskController.client.cancel(scatter_priors)

    return result_list


def bbsr_regress_dask(X, Y, pp_mat, weights_mat, G, genes, nS):
    """
    Execute regression (BBSR)

    :return: list
        Returns a list of regression results that the pileup_data can process
    """
    assert MPControl.is_dask()

    from inferelator.regression import bayes_stats
    DaskController = MPControl.client

    def regression_maker(j, x, y, pp, weights):
        level = 0 if j % 100 == 0 else 2
        utils.Debug.allprint(base_regression.PROGRESS_STR.format(gn=genes[j], i=j, total=G), level=level)
        data = bayes_stats.bbsr(x, utils.scale_vector(y), pp[j, :].flatten(), weights[j, :].flatten(), nS)
        data['ind'] = j
        return j, data

    # Scatter common data to workers
    [scatter_x] = DaskController.client.scatter([X.values], broadcast=True, hash=False)
    [scatter_pp] = DaskController.client.scatter([pp_mat.values], broadcast=True, hash=False)
    [scatter_weights] = DaskController.client.scatter([weights_mat.values], broadcast=True, hash=False)

    # Wait for scattering to finish before creating futures
    distributed.wait(scatter_x, timeout=DASK_SCATTER_TIMEOUT)
    distributed.wait(scatter_pp, timeout=DASK_SCATTER_TIMEOUT)
    distributed.wait(scatter_weights, timeout=DASK_SCATTER_TIMEOUT)

    future_list = [DaskController.client.submit(regression_maker, i, scatter_x,
                                                Y.get_gene_data(i, force_dense=True).flatten(),
                                                scatter_pp, scatter_weights)
                   for i in range(G)]

    # Collect results as they finish instead of waiting for all workers to be done
    result_list = process_futures_into_list(future_list)

    DaskController.client.cancel(scatter_x)
    DaskController.client.cancel(scatter_pp)
    DaskController.client.cancel(scatter_weights)

    return result_list


def sklearn_regress_dask(X, Y, model, G, genes, min_coef):
    """
    Execute regression (SKLearn)

    :return: list
        Returns a list of regression results that the pileup_data can process
    """
    assert MPControl.is_dask()

    from inferelator.regression import sklearn_regression
    DaskController = MPControl.client

    def regression_maker(j, x, y):
        level = 0 if j % 100 == 0 else 2
        utils.Debug.allprint(base_regression.PROGRESS_STR.format(gn=genes[j], i=j, total=G), level=level)
        data = sklearn_regression.sklearn_gene(x, utils.scale_vector(y), copy.copy(model))
        data['ind'] = j
        return j, data

    # Scatter common data to workers
    [scatter_x] = DaskController.client.scatter([X.values], broadcast=True, hash=False)

    # Wait for scattering to finish before creating futures
    distributed.wait(scatter_x, timeout=DASK_SCATTER_TIMEOUT)

    future_list = [DaskController.client.submit(regression_maker, i, scatter_x,
                                                Y.get_gene_data(i, force_dense=True).flatten())
                   for i in range(G)]

    # Collect results as they finish instead of waiting for all workers to be done
    result_list = process_futures_into_list(future_list)

    DaskController.client.cancel(scatter_x)

    return result_list


def lasso_stars_regress_dask(X, Y, alphas, num_subsamples, random_seed, method, params, G, genes):
    """
    Execute regression (LASSO-StARS)

    :return: list
        Returns a list of regression results that the pileup_data can process
    """
    assert MPControl.is_dask()

    from inferelator.regression import stability_selection
    DaskController = MPControl.client

    def regression_maker(j, x, y):
        level = 0 if j % 100 == 0 else 2
        utils.Debug.allprint(base_regression.PROGRESS_STR.format(gn=genes[j], i=j, total=G), level=level)
        data = stability_selection.stars_model_select(x, utils.scale_vector(y), alphas, num_subsamples=num_subsamples,
                                                      method=method, random_seed=random_seed, **params)
        data['ind'] = j
        return j, data

    # Scatter common data to workers
    [scatter_x] = DaskController.client.scatter([X.values], broadcast=True, hash=False)

    # Wait for scattering to finish before creating futures
    distributed.wait(scatter_x, timeout=DASK_SCATTER_TIMEOUT)

    future_list = [DaskController.client.submit(regression_maker, i, scatter_x,
                                                Y.get_gene_data(i, force_dense=True).flatten())
                   for i in range(G)]

    # Collect results as they finish instead of waiting for all workers to be done
    result_list = process_futures_into_list(future_list)

    DaskController.client.cancel(scatter_x)

    return result_list


def build_mi_array_dask(X, Y, bins, logtype):
    """
    Calculate MI into an array with dask (the naive map is very inefficient)

    :param X: np.ndarray (n x m1)
        Discrete array of bins
    :param Y: np.ndarray (n x m2)
        Discrete array of bins
    :param bins: int
        The total number of bins that were used to make the arrays discrete
    :param logtype: np.log func
        Which log function to use (log2 gives bits, ln gives nats)
    :return mi: np.ndarray (m1 x m2)
        Returns the mutual information array
    """

    assert MPControl.is_dask()

    from inferelator.regression.mi import _calc_mi, _make_table, _make_discrete

    # Get a reference to the Dask controller
    DaskController = MPControl.client

    m1, m2 = X.shape[1], Y.shape[1]

    def mi_make(i, x, y):
        x = _make_discrete(x, bins)
        return i, [_calc_mi(_make_table(x, y[:, j], bins), logtype=logtype) for j in range(m2)]

    # Scatter Y to workers and keep track as Futures
    [scatter_y] = DaskController.client.scatter([Y], broadcast=True, hash=False)

    # Wait for scattering to finish before creating futures
    distributed.wait(scatter_y, timeout=DASK_SCATTER_TIMEOUT)

    # Build an asynchronous list of Futures for each calculation of mi_make
    future_list = [DaskController.client.submit(mi_make, i,
                                                X[:, i].A.flatten() if sps.isspmatrix(X) else X[:, i].flatten(),
                                                scatter_y)
                   for i in range(m1)]

    # Collect results as they finish instead of waiting for all workers to be done
    mi_list = process_futures_into_list(future_list)

    # Convert the list of lists to an array
    mi = np.array(mi_list)
    assert (m1, m2) == mi.shape, "Array {sh} produced [({m1}, {m2}) expected]".format(sh=mi.shape, m1=m1, m2=m2)

    DaskController.client.cancel(scatter_y)

    return mi


def dask_map(func, *args, **kwargs):
    """
    Dask map

    :param func: function to map
    :type func: callable
    :param args: positional arguments for func
    :type args: iterable
    :param kwargs: keyword (non-iterable) arguments for func. Keywords will be passed to dask client.map.
    :return: List of mapped results
    :rtype: list
    """

    assert MPControl.is_dask()

    def _func_caller(f, i, *a, **k):
        return i, f(*a, **k)

    return process_futures_into_list([MPControl.client.client.submit(_func_caller, func, i, *za, **kwargs)
                                      for i, za in enumerate(zip(*args))])


def process_futures_into_list(future_list, raise_on_error=True, check_results=True):
    """
    Take a list of futures and turn them into a list of results
    Results must be of the form i, data (where i is the output order)

    :param future_list: A list of executing futures
    :type future_list: list
    :param raise_on_error: Should an error be raised if a job can't be restarted or just move on from it.
    Defaults to True
    :type raise_on_error: bool
    :param check_results: Should the result object be checked (and restarted if there's a problem)
    If False, this will raise an error with the result of a failed future is retrieved.
    Defaults to True.
    :type check_results: bool
    :return output_list: A list of results from the completed futures
    :rtype: list
    """

    DaskController = MPControl.client
    output_list = [None] * len(future_list)
    complete_gen = distributed.as_completed(future_list)

    for finished_future in complete_gen:

        DaskController.check_cluster_state()

        # Jobs can be cancelled in certain situations
        if check_results and (finished_future.cancelled() or (finished_future.status == "erred")):
            error = finished_future.exception()
            utils.Debug.vprint("Restarting job (Error: {er})".format(er=error), level=0)

            # Restart cancelled futures and put them back into the work pile
            try:
                DaskController.client.retry(finished_future)
                complete_gen.update([finished_future])
            except KeyError:
                if raise_on_error:
                    raise

        # In the event of success, get the data
        else:
            i, result_data = finished_future.result()
            output_list[i] = result_data
            finished_future.cancel()

    return output_list




import pandas as pd
import numpy as np
import os

from inferelator import utils
from inferelator.utils import Validator as check
from inferelator.postprocessing import results_processor
from inferelator.postprocessing import BETA_SIGN_COLUMN, MEDIAN_EXPLAIN_VAR_COLUMN


class ResultsProcessorMultiTask(results_processor.ResultsProcessor):
    """
    This results processor should handle the results of the MultiTask inferelator

    It will output the results for each task, as well as rank-combining to construct a network from all tasks
    """

    write_task_files = True

    tasks_names = None
    tasks_networks = None

    def __init__(self, betas, rescaled_betas, threshold=None, filter_method=None, metric=None):
        """
        :param betas: list(list(pd.DataFrame[G x K]) [B]) [T]
            A list of the task inferelator outputs per bootstrap per task
        :param rescaled_betas: list(list(pd.DataFrame[G x K]) [B]) [T]
            A list of the variance explained by each parameter per bootstrap per task
        :param threshold: float
            The proportion of bootstraps which an model weight must be non-zero for inclusion in the network output
        :param filter_method: str
            How to handle gold standard filtering ('overlap' filters to beta, 'keep_all_gold_standard' doesn't filter)
        :param metric: str / RankSummingMetric
            The scoring metric to use
        """

        super(ResultsProcessorMultiTask, self).__init__(betas, rescaled_betas, threshold, filter_method, metric)

        # Make up some default names
        # The workflow will have to replace these with real names if necessary
        self.tasks_names = list(map(str, range(len(self.betas))))

    @staticmethod
    def validate_init_args(betas, rescaled_betas, threshold=None, filter_method=None, metric=None):
        assert check.argument_type(betas, list)
        assert check.argument_list_type(betas, list)
        assert check.argument_list_type(betas[0], pd.DataFrame)
        assert check.argument_type(rescaled_betas, list)
        assert check.argument_list_type(rescaled_betas, list)
        assert check.argument_list_type(rescaled_betas[0], pd.DataFrame)
        assert all([check.dataframes_align(b_task + bresc_task) for b_task, bresc_task in zip(betas, rescaled_betas)])
        assert check.argument_enum(filter_method, results_processor.FILTER_METHODS, allow_none=True)
        assert check.argument_numeric(threshold, 0, 1, allow_none=True)

    def summarize_network(self, output_dir, gold_standard, priors):
        """
        Take the betas and rescaled beta_errors, construct a network, and test it against the gold standard
        :param output_dir: str
            Path to write files into. Don't write anything if this is None.
        :param gold_standard: pd.DataFrame [G x K]
            Gold standard to test the network against
        :param priors: list(pd.DataFrame [G x K])
            Prior data
        :return overall_result: InferelatorResult
            Returns an InferelatorResult for the final aggregate network
        """

        assert check.argument_type(priors, list)
        assert len(priors) == len(self.tasks_names)
        assert len(self.betas) == len(self.tasks_names)
        assert len(self.rescaled_betas) == len(self.tasks_names)

        # Create empty lists for task-specific results
        overall_confidences = []
        overall_resc_betas = []

        # Get intersection of indices
        gene_set = list(set([i for df in self.betas for i in df[0].index.tolist()]))
        tf_set = list(set([i for df in self.betas for i in df[0].columns.tolist()]))

        # Use the existing indices if there's no difference from the intersection
        gene_set = gene_set if len(self.betas[0][0].index.symmetric_difference(gene_set)) != 0 else self.betas[0][0].index
        tf_set = tf_set if len(self.betas[0][0].columns.symmetric_difference(tf_set)) != 0 else self.betas[0][0].columns

        # Create empty dataframes for task-specific results
        overall_sign = pd.DataFrame(np.zeros((len(gene_set), len(tf_set))),
                                    index=gene_set,
                                    columns=tf_set)

        overall_threshold = overall_sign.copy()

        if not isinstance(gold_standard, list):
            gold_standard = [gold_standard] * len(self.tasks_names)

        # Run the result processing on each individual task
        # Keep the confidences from the tasks so that a final aggregate network can be assembled
        # Store the individual task network results in a dict
        self.tasks_networks = {}
        for task_id, task_name in enumerate(self.tasks_names):
            task_rs_calc = self.metric(self.rescaled_betas[task_id], gold_standard[task_id],
                                       filter_method=self.filter_method)

            task_threshold, task_sign, task_nonzero = self.threshold_and_summarize(self.betas[task_id], self.threshold)
            task_resc_betas_mean, task_resc_betas_median = self.mean_and_median(self.rescaled_betas[task_id])

            task_extra_cols = {BETA_SIGN_COLUMN: task_sign, MEDIAN_EXPLAIN_VAR_COLUMN: task_resc_betas_median}

            task_network_data = self.process_network(task_rs_calc, priors[task_id], beta_threshold=task_threshold,
                                                     extra_columns=task_extra_cols)

            # Pile up data
            overall_confidences.append(_df_resizer(task_rs_calc.all_confidences, gene_set, tf_set))
            overall_resc_betas.append(_df_resizer(task_resc_betas_median, gene_set, tf_set))
            overall_sign += np.sign(_df_resizer(task_sign, gene_set, tf_set))
            overall_threshold += _df_resizer(task_threshold, gene_set, tf_set)

            m_name, score = task_rs_calc.score()
            utils.Debug.vprint("Task {t} Model {m}:\t{score}".format(t=task_name, m=m_name, score=score), level=0)

            task_result = self.result_object(task_network_data, task_threshold, task_rs_calc.all_confidences,
                                             task_rs_calc, betas_sign=task_sign, betas=self.betas[task_id])

            if self.write_task_files is True and output_dir is not None:
                task_result.write_result_files(os.path.join(output_dir, task_name))

            self.tasks_networks[task_id] = task_result

        overall_rs_calc = self.metric(overall_confidences, gold_standard[0], filter_method=self.filter_method)

        overall_threshold = (overall_threshold / len(overall_confidences) > self.threshold).astype(int)
        overall_resc_betas_mean, overall_resc_betas_median = self.mean_and_median(overall_resc_betas)
        extra_cols = {BETA_SIGN_COLUMN: overall_sign, MEDIAN_EXPLAIN_VAR_COLUMN: overall_resc_betas_median}

        m_name, score = overall_rs_calc.score()
        utils.Debug.vprint("Aggregate Model {m}:\t{score}".format(m=m_name, score=score), level=0)

        network_data = self.process_network(overall_rs_calc, None, beta_threshold=overall_threshold,
                                            extra_columns=extra_cols)

        overall_result = self.result_object(network_data, overall_threshold,
                                            _df_resizer(overall_rs_calc.all_confidences, gene_set, tf_set),
                                            overall_rs_calc, betas_sign=overall_sign)
        overall_result.write_result_files(output_dir)
        overall_result.tasks = self.tasks_networks

        return overall_result


def _df_resizer(df, row_labs, col_labs, fill=0):
    return df.reindex(row_labs, axis=0).reindex(col_labs, axis=1).fillna(fill)



"""
Implementation for the Transcription Factor Activity (TFA) based Inferelator workflow.
This workflow also has a design driver which will incorporate timecourse data.
This is the standard workflow for most applications.
"""

import numpy as np
from inferelator import workflow
from inferelator.preprocessing import design_response_translation
from inferelator.preprocessing.tfa import TFA, NoTFA
from inferelator.utils import InferelatorDataLoader, Debug, InferelatorData, Validator as check


class TFAWorkFlow(workflow.WorkflowBase):
    """
    TFAWorkFlow runs the timecourse driver and the TFA driver prior to regression.
    """
    # Design/response parameters
    delTmin = 0
    delTmax = 120
    tau = 45

    # Regression data
    # InferelatorData [N x G]
    design = None

    # InferelatorData [N x K]
    response = None
    half_tau_response = None

    # TFA implementation
    tfa_driver = TFA
    _tfa_output_file = None

    # Precalculated TFA
    _tfa_input_file = None
    _tfa_input_file_type = None

    # Design-Response Driver implementation
    drd_driver = design_response_translation.PythonDRDriver

    def set_design_settings(self, timecourse_response_driver=True, delTmin=None, delTmax=None, tau=None):
        """
        Set the parameters used in the timecourse design-response driver.

        :param timecourse_response_driver: A flag to indicate that the timecourse calculations should be performed.
            If set False, no other timecourse settings will have any effect.
            Defaults to True.
        :type timecourse_response_driver: bool
        :param delTmin: The minimum allowed time difference between timepoints to model as a time series. Provide in the
            same units as the metadata time column (usually minutes).
            Defaults to 0.
        :type delTmin: int, float
        :param delTmax: The maximum allowed time difference between timepoints to model as a time series. Provide in the
            same units as the metadata time column (usually minutes).
            Defaults to 120.
        :type delTmax: int, float
        :param tau: The tau parameter. Provide in the same units as the metadata time column (usually minutes).
            Defaults to 45.
        :type tau: int, float
        """

        if timecourse_response_driver is None:
            pass
        elif timecourse_response_driver:
            self.drd_driver = design_response_translation.PythonDRDriver
        else:
            self.drd_driver = None

        self._set_without_warning("delTmin", delTmin)
        self._set_without_warning("delTmax", delTmax)
        self._set_without_warning("tau", tau)

    def set_tfa(self, tfa_driver=None, tfa_output_file=None, tfa_input_file=None, tfa_input_file_type=None):
        """
        Perform or skip the TFA calculations; by default the design matrix will be transcription factor activity.
        If this is called with `tfa_driver = False`, the design matrix will be transcription factor expression.
        It is not necessary to call this function unless setting `tfa_driver = False`.

        :param tfa_driver: A flag to indicate that the TFA calculations should be performed.
            Defaults to True
        :type tfa_driver: bool
        :param tfa_output_file: A path to a TSV file which will be created with the calculated TFAs. Note that this file
            may contain TF expression if the TFA cannot be calculated for that TF.
            If None, no output file will be produced.
            Defaults to None
        :type tfa_output_file: str, optional
        :param tfa_input_file: A path to a TFA file which will be loaded and used in place of activity calculations
            If set, all TFA-related settings will be irrelevant. TSV file MUST be Samples X TFA.
            If None, the inferelator will calculate TFA
            Defaults to None
        :type tfa_output_file: str, optional
        :param tfa_input_file_type: A string which identifies file type. Accepts "tsv" and "h5ad".
            If None, assume the file is a TSV
            Defaults to None
        :type tfa_output_file: str, optional
        """

        if tfa_driver is None:
            pass
        elif tfa_driver:
            self.tfa_driver = TFA
        else:
            self.tfa_driver = NoTFA

        self._set_with_warning("_tfa_output_file", tfa_output_file)
        self._set_file_name("_tfa_input_file", tfa_input_file)
        self._set_without_warning("_tfa_input_file_type", tfa_input_file_type)

    def run(self):
        """
        Execute workflow, after all configuration.
        """

        # Set the random seed (for bootstrap selection)
        np.random.seed(self.random_seed)

        # Call the startup workflow
        self.startup()

        # Run regression after startup
        betas, rescaled_betas = self.run_regression()

        # Write the results out to a file
        return self.emit_results(betas, rescaled_betas, self.gold_standard, self.priors_data)

    def startup_run(self):
        self.get_data()
        self.process_priors_and_gold_standard()

    def startup_finish(self):
        self.align_priors_and_expression()

        self.compute_common_data()
        self.compute_activity()

        # Most operations will be column-wise; change sparse type if needed here
        self.response.to_csc()

    def run_regression(self):
        raise NotImplementedError

    def run_bootstrap(self, bootstrap):
        raise NotImplementedError

    def compute_activity(self):
        """
        Compute Transcription Factor Activity
        """
        # If there is a tfa driver, run it to calculate TFA from the prior & expression data

        if self._tfa_input_file is not None:
            self.load_activity()
        else:
            self._recalculate_design()

        self.half_tau_response = None

        if self._tfa_output_file is not None and self.is_master():
            self.create_output_dir()
            self.design.to_csv(self.output_path(self._tfa_output_file), sep="\t")

    def _recalculate_design(self):
        """
        Use the TFA driver to recalculate the design matrix
        """
        self.design.convert_to_float()
        self.half_tau_response.convert_to_float()
        self.design = self.tfa_driver().compute_transcription_factor_activity(self.priors_data,
                                                                              self.design,
                                                                              self.half_tau_response)
        Debug.vprint("Rebuilt design matrix {d} with TF activity".format(d=self.design.shape), level=1)

    def load_activity(self, file=None, file_type=None):

        file = self._tfa_input_file if file is None else file
        file_type = self._tfa_input_file_type if file_type is None else file_type

        loader = InferelatorDataLoader(input_dir=self.input_dir, file_format_settings=self._file_format_settings)

        if file_type.lower() == "h5ad":
            self.design = loader.load_data_h5ad(file)
        elif self._expression_loader.lower() == "tsv":
            self.design = loader.load_data_tsv(file)

        Debug.vprint("Loaded {f} as design matrix {d}".format(d=self.design.shape, f=file), level=1)

        self.design.trim_genes(remove_constant_genes=False,
                               trim_gene_list=self.design.gene_names.intersection(self.tf_names))

        Debug.vprint("Trimmed to {d} for TF activity".format(d=self.design.shape, f=file), level=1)

        assert check.indexes_align([self.design.sample_names, self.response.sample_names])

    def emit_results(self, betas, rescaled_betas, gold_standard, priors):
        """
        Output result report(s) for workflow run.
        """
        if self.is_master():
            self.create_output_dir()
            rp = self._result_processor_driver(betas, rescaled_betas, filter_method=self.gold_standard_filter_method,
                                               metric=self.metric)
            self.results = rp.summarize_network(self.output_dir, gold_standard, priors)
            return self.results
        else:
            return None

    def compute_common_data(self):
        """
        Compute common data structures like design and response matrices.
        """

        drd = self.drd_driver(metadata_handler=self.metadata_handler,
                              return_half_tau=True) if self.drd_driver is not None else None

        # If there is no design-response driver set, use the expression data for design and response
        # Also do this if there is no usable metadata
        if drd is None or not drd.validate_run(self.data.meta_data):
            self.design, self.response, self.half_tau_response = self.data, self.data, self.data

        # Otherwise calculate the design-response ODE
        # TODO: Rewrite DRD for InferelatorData
        # TODO: This is *horrifying* as is from a memory perspective
        # TODO: Really fix this soon
        else:
            Debug.vprint('Creating design and response matrix ... ')
            drd.delTmin, drd.delTmax, drd.tau = self.delTmin, self.delTmax, self.tau

            design, response, half_tau_response = drd.run(self.data.to_df().T, self.data.meta_data)
            self.design = InferelatorData(design.T)
            self.response = InferelatorData(response.T)
            self.half_tau_response = InferelatorData(half_tau_response.T)

        Debug.vprint("Constructed design {d} and response {r} matrices".format(d=self.design.shape,
                                                                               r=self.response.shape), level=1)

        self.data = None



import numpy as np
from scipy import linalg, sparse
from inferelator import utils


class TFA:
    """ TFA calculates transcription factor activity using matrix pseudoinverse """

    def compute_transcription_factor_activity(self, prior, expression_data, expression_data_halftau=None,
                                              keep_self=False):
        """
        Calculate TFA from a prior and expression data object

        :param prior: pd.DataFrame [G x K]
        :param expression_data: InferelatorData [N x G]
        :param expression_data_halftau: InferelatorData [N x G]
        :param keep_self: bool
        :return: InferelatorData [N x K]
        """

        prior, activity_tfs, expr_tfs = self._check_prior(prior, expression_data, keep_self=keep_self)

        activity = np.zeros((expression_data.shape[0], prior.shape[1]), dtype=np.float64)

        if len(activity_tfs) > 0:
            a_cols = prior.columns.isin(activity_tfs)
            expr = expression_data_halftau if expression_data_halftau is not None else expression_data
            activity[:, a_cols] = self._calculate_activity(prior.loc[:, activity_tfs].values, expr)

        if len(expr_tfs) > 0:
            activity[:, prior.columns.isin(expr_tfs)] = expression_data.get_gene_data(expr_tfs, force_dense=True)

        data_name = "Activity" if expression_data.name is None else "{n} Activity".format(n=expression_data.name)
        return utils.InferelatorData(activity,
                                     gene_names=prior.columns,
                                     sample_names=expression_data.sample_names,
                                     meta_data=expression_data.meta_data,
                                     name=data_name)

    def _check_prior(self, prior, expression_data, keep_self=False):
        if not keep_self:
            prior = utils.df_set_diag(prior, 0)

        activity_tfs, expr_tfs, drop_tfs = self._determine_tf_status(prior, expression_data)

        if len(drop_tfs) > 0:
            msg = "{n} TFs are removed from activity (no expression or prior exists)".format(n=len(drop_tfs))
            utils.Debug.vprint(msg, level=0)
            utils.Debug.vprint(" ".join(drop_tfs), level=1)

        prior = prior.drop(drop_tfs, axis=1)

        return prior, activity_tfs, expr_tfs

    @staticmethod
    def _determine_tf_status(prior, expression_data):

        # These TFs can have activity calculations performed for them because there are priors
        activity_tfs = ((prior != 0).sum(axis=0) != 0).values
        # These TFs match gene expression only (no activity calculation)
        expr_tfs = prior.columns.isin(expression_data.gene_names)
        expr_tfs &= ~activity_tfs

        return prior.columns[activity_tfs], prior.columns[expr_tfs], prior.columns[~(activity_tfs | expr_tfs)]

    @staticmethod
    def _calculate_activity(prior, expression_data):

        prior_dtype = np.float32 if expression_data.values.dtype == np.float32 else np.float64
        return utils.DotProduct.dot(expression_data.values, sparse.csr_matrix(linalg.pinv2(prior).T, dtype=prior_dtype),
                                    dense=True, cast=True)


class NoTFA(TFA):
    """ NoTFA creates an activity matrix from the expression data only """

    def compute_transcription_factor_activity(self, prior, expression_data, expression_data_halftau=None, keep_self=False):
        utils.Debug.vprint("Setting Activity to Expression Values", level=1)
        tf_gene_overlap = prior.columns[prior.columns.isin(expression_data.gene_names)]

        data_name = "Activity" if expression_data.name is None else "{n} Activity".format(n=expression_data.name)
        return utils.InferelatorData(expression_data.get_gene_data(tf_gene_overlap, copy=True, force_dense=True),
                                     sample_names=expression_data.sample_names,
                                     meta_data=expression_data.meta_data,
                                     gene_names=tf_gene_overlap,
                                     name=data_name)



from __future__ import division

import numpy as np
import itertools
import math
import warnings
import scipy.special
import scipy.linalg

from inferelator import utils
from inferelator.regression import base_regression


def bbsr(X, y, pp, weights, max_k, ordinary_least_squares=False):
    """
    Run BBSR to regress a response variable y in n conditions against predictors X in n conditions. Use the prior
    predictors matrix to filter the number of predictors from something massive to max_k.
    :param X: np.ndarray [N x K]
        Predictor features
    :param y: np.ndarray [N,]
        Response variables
    :param pp: np.ndarray [K,]
        Predictors to model with
    :param weights: np.ndarray [K,]
        Weight matrix
    :param max_k: int
        Max number of predictors
    :return: dict
        pp: Boolean array indicating which predictors are included in the model                 [K,]
        betas: Float array indicating the beta for each predictor included in the model         [K,]
        betas_resc: Float array indicating how much each predictor is contributing to the model [K,]
    """

    # Skip everything if there are no predictors in pp or if variance of y is 0
    if (pp.sum() == 0) or (np.var(y) == 0) or (np.all(~np.isfinite(y))):
        return dict(pp=np.repeat(True, pp.shape[0]).tolist(),
                    betas=np.zeros(pp.shape[0]),
                    betas_resc=np.zeros(pp.shape[0]))

    # Subset data to desired predictors
    pp_idx = base_regression.bool_to_index(pp)
    utils.Debug.vprint("Beginning regression with {pp_len} predictors".format(pp_len=len(pp_idx)), level=2)

    x = X[:, pp_idx]
    gprior = weights[pp_idx].astype(np.dtype(float))

    # Make sure arrays are 2d
    utils.make_array_2d(x)
    utils.make_array_2d(y)
    utils.make_array_2d(gprior)

    # Reduce predictors to max_k
    pp[pp_idx] = reduce_predictors(x, y, gprior, max_k, ordinary_least_squares=ordinary_least_squares)
    pp_idx = base_regression.bool_to_index(pp)

    utils.Debug.vprint("Reduced to {pp_len} predictors".format(pp_len=len(pp_idx)), level=2)

    # Check and make sure that there are still some predictors left
    if pp.sum() == 0:
        return dict(pp=np.repeat(True, pp.shape[0]).tolist(),
                    betas=np.zeros(pp.shape[0]),
                    betas_resc=np.zeros(pp.shape[0]))

    # Resubset with the newly reduced predictors
    x = X[:, pp_idx]
    gprior = weights[pp_idx].astype(np.dtype(float))
    utils.make_array_2d(gprior)

    betas = best_subset_regression(x, y, gprior, ordinary_least_squares=ordinary_least_squares)
    betas_resc = base_regression.predict_error_reduction(x, y, betas)

    return dict(pp=pp,
                betas=betas,
                betas_resc=betas_resc)


def best_subset_regression(x, y, gprior, ordinary_least_squares=False):
    """

    :param x: np.ndarray
        Independent (predictor) variables [n x k]
    :param y: np.ndarray
        Dependent (response) variable [n x 1]
    :param gprior: np.ndarray
        Weighted priors [k x 1]
    :return:
    """
    (n, k) = x.shape
    combos = combo_index(k)

    bic_combos = calc_all_expected_BIC(x, y, gprior, combos, check_rank=False,
                                       ordinary_least_squares=ordinary_least_squares)

    best_betas = np.zeros(k, dtype=np.dtype(float))
    try:
        best_combo = combos[:, _best_combo_idx(x, bic_combos, combos)]
    except np.linalg.LinAlgError:
        return best_betas

    if best_combo.sum() > 0:
        best_betas = base_regression.recalculate_betas_from_selected(x, y, best_combo)

    return best_betas


def reduce_predictors(x, y, gprior, max_k, ordinary_least_squares=False):
    """
    Determine which predictors are the most valuable by calculating BICs for single and pairwise predictor models
    :param x: np.ndarray [n x k]
    :param y: np.ndarray [n x 1]
    :param gprior: [k x 1]
    :param max_k: int
    :return: np.ndarray [k,]
    """
    (_, k) = x.shape

    if k <= max_k:
        return np.ones(k, dtype=np.dtype(bool))
    else:
        # Get BIC for every combination of single or double predictors
        combos = np.hstack((np.diag(np.repeat(True, k)), select_index(k)))
        bic = calc_all_expected_BIC(x, y, gprior, combos, ordinary_least_squares=ordinary_least_squares)

        reset = np.seterr(divide='ignore', invalid='ignore')
        bic = np.multiply(combos.T, bic.reshape(-1, 1)).sum(axis=0)

        # Return a boolean index pointing to the lowest BIC predictors
        predictors = np.zeros(k, dtype=np.dtype(bool))
        predictors[np.argsort(bic)[0:max_k]] = True

        np.seterr(**reset)
        return predictors


def calc_all_expected_BIC(x, y, g, combinations, check_rank=True, ordinary_least_squares=False):
    """
    Calculate BICs for every combination of predictors given in combinations
    :param x: np.ndarray [n x k]
        Array of predictor data
    :param y: np.ndarray [n x 1]
        Array of response data
    :param g: np.ndarray [k x 1]
        Weights for predictors
    :param combinations: np.ndarray [k x c]
        Combinations of predictors to try; each combination should be booleans with a length corresponding to the
        number of predictors
    :param check_rank: bool
        Explicitly check to see that xTx is nonsingular for every combination. If false, will only catch singular xTx
        that causes np.linalg.solve to throw an exception
    :return: np.ndarray [c,]
        Array of BICs corresponding to each combination
    """
    (n, k) = x.shape
    c = combinations.shape[1]
    utils.make_array_2d(g)

    # Sanity check the data
    assert n == y.shape[0]
    assert k == combinations.shape[0]

    # Precalculate xTx and xTy
    digamma_shape = scipy.special.digamma(n / 2.0)
    xtx = np.dot(x.T, x)  # [k x k]
    xty = np.dot(x.T, y)  # [k x 1]

    # Calculate the g-prior
    gprior = np.repeat(np.sqrt(1.0 / (g + 1.0)), k, axis=1)
    gprior = np.multiply(gprior, gprior.T)
    bic = np.zeros(c, dtype=np.dtype(float))

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        for i in range(c):
            # Convert the boolean slice into an index
            c_idx = base_regression.bool_to_index(combinations[:, i])
            k_included = len(c_idx)

            # Check for a null model
            if k_included == 0:
                bic[i] = n * np.log(np.var(y, ddof=1))
                continue

            # Calculate the rate parameter from this specific combination of predictors
            try:
                xtx_slice = xtx[:, c_idx][c_idx, :]

                model_beta = _solve_model(xtx_slice, xty[c_idx], check_rank=check_rank)
                model_ssr = ssr(x[:, c_idx], y, model_beta)
                if ordinary_least_squares:
                    bic[i] = _calc_BIC_RSS(n, k_included, model_ssr)
                else:
                    scale_param = _calc_ig_scale(model_beta, model_ssr, xtx_slice, gprior[:, c_idx][c_idx, :])
                    if np.isfinite(scale_param) and scale_param > 0:
                        bic[i] = _calc_BIC_inverse_gamma(n, k_included, digamma_shape, scale_param)
                    else:
                        raise np.linalg.LinAlgError
            except np.linalg.LinAlgError:
                bic[i] = np.inf

    return bic


def _calc_BIC_inverse_gamma(n, k, shape, scale):
    return n * (np.log(scale) - shape) + k * np.log(n)


def _calc_BIC_RSS(n, k, model_ssr):
    if model_ssr <= 0:
        model_ssr = np.finfo(float).eps

    return n * (np.log(model_ssr / n)) + k * np.log(n)


def _calc_ig_scale(beta_hat, model_ssr, xtx, gprior):
    # Calculate the rate parameter with a g-prior
    beta_flip = (0 - beta_hat.T)
    rate = xtx * gprior
    rate = np.dot(rate, beta_flip.T)
    rate = np.dot(beta_flip, rate)

    # Return the mean of the SSR and the rate parameter
    return (model_ssr + rate) / 2


def _solve_model(xtx, xty, check_rank=True):
    # Check to see if xTx is nonsingular (if necessary)
    if check_rank and not _matrix_full_rank(xtx):
        raise np.linalg.LinAlgError

    # Solve xTx against xTy
    return scipy.linalg.solve(xtx, xty, assume_a='sym')


def _best_combo_idx(x, bic, combo):
    """
    Find the lowest BIC combination that comes from a nonsingular xTx
    :param x: [n x k]
    :param bic: [c,] array of floats
    :param combo: [k x c]
    :return:
    """

    for i in range(combo.shape[1]):
        bic_idx = np.argmin(bic)  # In case of a tie, np.argmin returns the leftmost index
        c = combo[:, bic_idx]

        if c.sum() == 0:
            return bic_idx

        x_slice = x[:, c]
        if _matrix_full_rank(np.dot(x_slice.T, x_slice)):
            return bic_idx
        else:
            bic[bic_idx] = np.inf

    raise np.linalg.LinAlgError


def _matrix_full_rank(mat, tol=1e-10):
    return np.linalg.matrix_rank(mat, tol=tol) == mat.shape[1]


def ssr(x, y, beta):
    """
    Sum of squared residuals (Y - XB)
    :param x: np.ndarray [N x M]
    :param y: np.ndarray [N x P]
    :param beta: np.ndarray [M x P]
    :return: float
    """
    assert x.shape[1] == beta.shape[0]

    resid = y - np.dot(x, beta)
    return (resid * resid).sum()


def combo_index(n):
    """
    Generate a boolean array that can mask to generate every possible combination of n objects
    :param n: int
        Number of objects
    :return: np.array
        [n x 2^n] array of bool
    """
    assert n >= 0

    return np.array(list(itertools.product([False, True], repeat=n))).T


def select_index(n, r=2):
    """
    Generate a boolean array that can mask to generate every selection of r objects from a total pool of n objects
    :param n: int
        Number of objects
    :param r: int
        Number of objects to select
    :return: np.arrap
        [n x n!/(r!(n-r)!)] array of bool
    """
    assert n >= 0

    combos = int(math.factorial(n) / (math.factorial(r) * math.factorial(n - r)))
    idx = np.array(list(itertools.combinations(range(n), r)))
    idx = idx + np.array(range(idx.shape[0])).reshape(-1, 1) * n

    arr = np.full(combos * n, False, dtype=np.dtype(bool))
    arr[idx.reshape(1, -1)] = True

    return arr.reshape((combos, n)).T



import numpy as np
import pandas as pd
from copy import deepcopy
from scipy.special import comb
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LinearRegression

from inferelator.distributed.inferelator_mp import MPControl
from inferelator import utils
from inferelator.utils import Validator as check
from inferelator import default
from inferelator.regression import base_regression

# Shadow built-in zip with itertools.izip if this is python2 (This puts out a memory dumpster fire)
try:
    from itertools import izip as zip
except ImportError:
    pass

DEFAULT_prior_weight = 1.0
DEFAULT_Cs = np.logspace(np.log10(0.01), np.log10(10), 20)[::-1]

MAX_ITER = 1000
TOL = 1e-2
MIN_WEIGHT_VAL = 0.1
MIN_RSS = 1e-10


def run_regression_EBIC(X, Y, TFs, tasks, gene, prior, Cs=None, Ss=None, lambda_Bs=None,
                        lambda_Ss=None, scale_data=False, return_lambdas=False):
    """
    Run multitask regression. Search the regularization coefficient space and select the model with the
    lowest eBIC.

    :param X: list(np.ndarray [N x K]) [t]
        List consisting of design matrixes for each task
    :param Y: list(np.ndarray [N x 1]) [t]
        List consisting of response matrixes for each task
    :param TFs: list [K]
        List of TF names for each feature
    :param tasks: list(int) [t]
        List identifying each task
    :param gene: str
        The gene being modeled
    :param prior: np.ndarray [K x T]
        The priors for this gene in a TF x Task array
    :return: dict
    """

    assert len(X) == len(Y)
    assert len(X) == len(tasks)

    # The number of tasks
    n_tasks = len(X)

    # The number of predictors
    n_preds = X[0].shape[1]

    # A list of the number of samples for each task
    n_samples = [X[k].shape[0] for k in range(n_tasks)]

    ###### EBIC ######

    # Create empty block and sparse matrixes
    sparse_matrix = np.zeros((n_preds, n_tasks))
    block_matrix = np.zeros((n_preds, n_tasks))

    # Set the starting EBIC to infinity
    min_ebic = float('Inf')
    model_output = None

    # Calculate covariances
    X = scale_list_of_arrays(X) if scale_data else X
    Y = scale_list_of_arrays(Y) if scale_data else Y
    cov_C, cov_D = _covariance_by_task(X, Y)

    # Calculate lambda_B defaults if not provided
    if lambda_Bs is None:

        # Start with a default lambda B based on tasks, predictors, and samples
        lambda_B_param = np.sqrt((n_tasks * np.log(n_preds)) / np.mean(n_samples))

        # Modify by multiplying against the values in Cs
        lambda_Bs = lambda_B_param * np.array(DEFAULT_Cs if Cs is None else Cs)

    # Iterate through lambda_Bs
    for b in lambda_Bs:

        # Set scaling values if not provided
        Ss = np.linspace((1.0/n_tasks)+0.01, 0.99, 10)[::-1] if Ss is None else Ss

        # Iterate through lambda_Ss or calculate lambda Ss based on heuristic
        for s in lambda_Ss if lambda_Ss is not None else b * np.array(Ss):

            # Fit model
            combined_weights, sparse_matrix, block_matrix = amusr_fit(cov_C, cov_D, b, s,
                                                                      sparse_matrix, block_matrix, prior)

            # Score model
            ebic_score = ebic(X, Y, combined_weights, n_tasks, n_samples, n_preds)

            # Keep the model if it's the lowest scoring
            if ebic_score < min_ebic:
                min_ebic = ebic_score
                model_output = combined_weights
                opt_b, opt_s = b, s

    ###### RESCALE WEIGHTS ######
    output = {}

    if model_output is not None:
        for kx, k in enumerate(tasks):
            nonzero = model_output[:,kx] != 0
            if nonzero.sum() > 0:
                cTFs = np.asarray(TFs)[model_output[:,kx] != 0]
                output[k] = _final_weights(X[kx][:, nonzero], Y[kx], cTFs, gene)

    return (output, opt_b, opt_s) if return_lambdas else output

class AMuSR_regression(base_regression.BaseRegression):

    X = None  # list(InferelatorData)
    Y = None  # list(InferelatorData)
    priors = None  # list(pd.DataFrame [G, K]) OR pd.DataFrame [G, K]
    tfs = None  # pd.Index OR list
    genes = None  # pd.Index OR list

    K = None  # int
    G = None  # int
    n_tasks = None  # int
    prior_weight = DEFAULT_prior_weight  # float
    remove_autoregulation = True  # bool

    lambda_Bs = None
    lambda_Ss = None
    Cs = None
    Ss = None

    regression_function = staticmethod(run_regression_EBIC)

    def __init__(self, X, Y, tfs=None, genes=None, priors=None, prior_weight=1, remove_autoregulation=True,
                 lambda_Bs=None, lambda_Ss=None, Cs=None, Ss=None):
        """
        Set up a regression object for multitask regression
        :param X: list(InferelatorData)
        :param Y: list(InferelatorData)
        :param priors: pd.DataFrame [G, K]
        :param prior_weight: float
        :param remove_autoregulation: bool
        :param lambda_Bs: list(float) [None]
        :param lambda_Ss: list(float) [None]
        :param Cs: list(float) [None] 
        :param Ss: list(float) [None]
        """

        # Check input types are correct
        assert check.argument_type(X, list)
        assert check.argument_type(Y, list)
        assert check.argument_type(tfs, (list, pd.Series, pd.Index), allow_none=True)
        assert check.argument_type(genes, (list, pd.Series, pd.Index), allow_none=True)
        assert check.argument_numeric(prior_weight)
        assert len(X) == len(Y)

        # Set the data into the regression object
        self.X = scale_list_of_data(X)
        self.Y = scale_list_of_data(Y)
        self.n_tasks = len(X)

        # Set the priors and weight into the regression object
        self.priors = priors
        self.prior_weight = float(prior_weight)

        # Construct a list of regulators if they are not passed in from the union of the task regulators
        if tfs is None:
            tfs = [design.gene_names for design in X]
            self.tfs = filter_genes_on_tasks(tfs, "union")
        else:
            self.tfs = tfs

        # Construct a list of genes if they are not passed in from the union of the task genes
        if genes is None:
            genes = [resp.gene_names for resp in Y]
            self.genes = filter_genes_on_tasks(genes, "union")
        else:
            self.genes = genes

        # Set the regulators and targets into the regression object
        self.K, self.G = len(tfs), len(genes)
        self.remove_autoregulation = remove_autoregulation

        # Set the regularization coefficients into the regression object
        self.lambda_Bs = lambda_Bs
        self.lambda_Ss = lambda_Ss
        self.Cs = Cs
        self.Ss = Ss

    def regress(self, regression_function=None):
        """
        Execute multitask (AMUSR)
        :return: list
            Returns a list of regression results that the amusr_regression pileup_data can process
        """

        regression_function = self.regression_function if regression_function is None else regression_function

        if MPControl.is_dask():
            from inferelator.distributed.dask_functions import amusr_regress_dask
            return amusr_regress_dask(self.X, self.Y, self.priors, self.prior_weight, self.n_tasks, self.genes,
                                      self.tfs, self.G, remove_autoregulation=self.remove_autoregulation,
                                      regression_function=regression_function)

        def regression_maker(j):
            level = 0 if j % 100 == 0 else 2
            utils.Debug.allprint(base_regression.PROGRESS_STR.format(gn=self.genes[j], i=j, total=self.G),
                                 level=level)

            gene = self.genes[j]
            x, y, tasks = [], [], []

            if self.remove_autoregulation:
                tfs = [t for t in self.tfs if t != gene]
            else:
                tfs = self.tfs

            for k in range(self.n_tasks):
                if gene in self.Y[k].gene_names:
                    x.append(self.X[k].get_gene_data(tfs))  # list([N, K])
                    y.append(self.Y[k].get_gene_data(gene, force_dense=True).reshape(-1, 1))  # list([N, 1])
                    tasks.append(k)  # [T,]

            prior = format_prior(self.priors, gene, tasks, self.prior_weight)
            return regression_function(x, y, tfs, tasks, gene, prior, Cs=self.Cs, Ss=self.Ss,
                                       lambda_Bs=self.lambda_Bs, lambda_Ss=self.lambda_Ss)

        return MPControl.map(regression_maker, range(self.G))

    def pileup_data(self, run_data):

        weights = []
        rescaled_weights = []

        for k in range(self.n_tasks):
            results_k = []
            for res in run_data:
                try:
                    results_k.append(res[k])
                except (KeyError, IndexError):
                    pass

            results_k = pd.concat(results_k)
            weights_k = _format_weights(results_k, 'weights', self.genes, self.tfs)
            rescaled_weights_k = _format_weights(results_k, 'resc_weights', self.genes, self.tfs)
            rescaled_weights_k[rescaled_weights_k < 0.] = 0

            weights.append(weights_k)
            rescaled_weights.append(rescaled_weights_k)

        return weights, rescaled_weights

def amusr_fit(cov_C, cov_D, lambda_B=0., lambda_S=0., sparse_matrix=None, block_matrix=None, prior=None,
              max_iter=MAX_ITER, tol=TOL, min_weight=MIN_WEIGHT_VAL):
    """
    Fits regression model in which the weights matrix W (predictors x tasks)
    is decomposed in two components: B that captures block structure across tasks
    and S that allows for the differences.
    reference: Jalali et al., NIPS 2010. A Dirty Model for Multi-task Learning.
    :param cov_C: np.ndarray [T x K]
        Covariance of the predictors K to the response gene by task
    :param cov_D: np.ndarray [T x K x K]
        Covariance of the predictors K to K by task
    :param lambda_B: float
        Penalty coefficient for the block matrix
    :param lambda_S: float
        Penalty coefficient for the sparse matrix
    :param sparse_matrix: np.ndarray [K x T]
        Matrix of model coefficients for each predictor by each task that are unique to each task
    :param block_matrix: np.ndarray [K x T]
        Matrix of model coefficients for each predictor by each task that are shared between each task
    :param prior: np.ndarray [T x K]
        Matrix of known prior information
    :param max_iter: int
        Maximum number of model iterations if tol is not reached
    :param tol: float
        The tolerance for the stopping criteria (convergence)
    :param min_weight: float
        Regularize any weights below this threshold to 0
    :return combined_weights: np.ndarray [K x T]
        Matrix of model coefficients for each predictor by each task that are the summation of both the task-specific
        model and the shared model
    :return sparse_matrix: np.ndarray [K x T]
        Matrix of model coefficients for each predictor by each task that are unique to each task
    :return block_matrix: np.ndarray [K x T]
        Matrix of model coefficients for each predictor by each task that are shared between each task
    """

    assert check.argument_type(lambda_B, (float, int, np.int64, np.float32))
    assert check.argument_type(lambda_S, (float, int, np.int64, np.float32))
    assert check.argument_type(max_iter, int)
    assert check.argument_type(tol, float)
    assert check.argument_type(min_weight, float)
    assert cov_C.shape[0] == cov_D.shape[0]
    assert cov_C.shape[1] == cov_D.shape[1]
    assert cov_D.shape[1] == cov_D.shape[2]

    n_tasks = cov_C.shape[0]
    n_features = cov_C.shape[1]

    # if S and B are provided -- warm starts -- will run faster
    if sparse_matrix is None or block_matrix is None:
        sparse_matrix = np.zeros((n_features, n_tasks))
        block_matrix = np.zeros((n_features, n_tasks))

    # If there is no prior for weights, create an array of 1s
    prior = np.ones((n_features, n_tasks)) if prior is None else prior

    # Initialize weights
    combined_weights = sparse_matrix + block_matrix
    for i in range(max_iter):

        # Keep a reference to the old combined_weights
        _combined_weights_old = combined_weights

        # Update sparse and block-sparse coefficients
        sparse_matrix = updateS(cov_C, cov_D, block_matrix, sparse_matrix, lambda_S, prior, n_tasks, n_features)
        block_matrix = updateB(cov_C, cov_D, block_matrix, sparse_matrix, lambda_B, prior, n_tasks, n_features)

        # Weights matrix (W) is the sum of a sparse (S) and a block-sparse (B) matrix
        combined_weights = sparse_matrix + block_matrix

        # If convergence tolerance reached, break loop and move on
        if np.max(np.abs(combined_weights - _combined_weights_old)) < tol:
            break

    # Set small values of W to zero
    # Since we don't run the algorithm until update equals zero
    combined_weights[np.abs(combined_weights) < min_weight] = 0

    return combined_weights, sparse_matrix, block_matrix

def updateS(C, D, B, S, lamS, prior, n_tasks, n_features):
    """
    returns updated coefficients for S (predictors x tasks)
    lasso regularized -- using cyclical coordinate descent and
    soft-thresholding
    """
    # update each task independently (shared penalty only)
    for k in range(n_tasks):
        c = C[k]; d = D[k]
        b = B[:,k]; s = S[:,k]
        p = prior[:,k]
        # cycle through predictors
        for j in range(n_features):
            # set sparse coefficient for predictor j to zero
            s_tmp = deepcopy(s)
            s_tmp[j] = 0.
            # calculate next coefficient based on fit only
            if d[j,j] == 0:
                alpha = 0
            else:
                alpha = (c[j]-np.sum((b+s_tmp)*d[j]))/d[j,j]
            # lasso regularization
            if abs(alpha) <= p[j]*lamS:
                s[j] = 0.
            else:
                s[j] = alpha-(np.sign(alpha)*p[j]*lamS)
        # update current task
        S[:,k] = s

    return(S)

def updateB(C, D, B, S, lamB, prior, n_tasks, n_features):
    """
    returns updated coefficients for B (predictors x tasks)
    block regularized (l_1/l_inf) -- using cyclical coordinate descent and
    soft-thresholding on the l_1 norm across tasks
    reference: Liu et al, ICML 2009. Blockwise coordinate descent procedures
    for the multi-task lasso, with applications to neural semantic basis discovery.
    """
    #p = prior.min(axis=1)
    # cycles through predictors
    for j in range(n_features):
        # initialize next coefficients
        alphas = np.zeros(n_tasks)
        # update tasks for each predictor together
        for k in range(n_tasks):
            # get task covariance update terms
            c = C[k]; d = D[k]
            # get previous block-sparse and sparse coefficients
            b = B[:,k]; s = S[:,k]
            # set block-sparse coefficient for feature j to zero
            b_tmp = deepcopy(b)
            b_tmp[j] = 0.
            # calculate next coefficient based on fit only
            if d[j,j] == 0:
                alphas[k] = 0
            else:
                alphas[k] = (c[j]-np.sum((b_tmp+s)*d[:,j]))/d[j,j]
        # set all tasks to zero if l1-norm less than lamB
        if np.linalg.norm(alphas, 1) <= lamB:
            B[j,:] = np.zeros(n_tasks)
        # regularized update for predictors with larger l1-norm
        else:
            # find number of coefficients that would make l1-norm greater than penalty
            indices = np.abs(alphas).argsort()[::-1]
            sorted_alphas = alphas[indices]
            m_star = np.argmax((np.abs(sorted_alphas).cumsum()-lamB)/(np.arange(n_tasks)+1))
            # initialize new weights
            new_weights = np.zeros(n_tasks)
            # keep small coefficients and regularize large ones (in above group)
            for k in range(n_tasks):
                idx = indices[k]
                if k > m_star:
                    new_weights[idx] = sorted_alphas[k]
                else:
                    sign = np.sign(sorted_alphas[k])
                    update_term = np.sum(np.abs(sorted_alphas)[:m_star+1])-lamB
                    new_weights[idx] = (sign/(m_star+1))*update_term
            # update current predictor
            B[j,:] = new_weights

    return(B)

def _covariance_by_task(X, Y):
    """
    Returns C and D, containing terms for covariance update for OLS fit
    C: transpose(X_j)*Y for each feature j
    D: transpose(X_j)*X_l for each feature j for each feature l
    Reference: Friedman, Hastie, Tibshirani, 2010 in Journal of Statistical Software
    Regularization Paths for Generalized Linear Models via Coordinate Descent
    :param X: list(np.ndarray [N x K]) [T]
        List of design values for each task. Must be aligned on the feature (K) axis.
    :param Y: list(np.ndarray [N x 1]) [T]
        List of response values for each task
    :return cov_C, cov_D: np.ndarray [T x K], np.ndarray [T x K x K]
        Covariance of the predictors K to the response gene by task
        Covariance of the predictors K to K by task
    """

    assert check.argument_type(X, list)
    assert check.argument_type(Y, list)
    assert len(X) == len(Y)
    assert max([xk.shape[1] for xk in X]) == min([xk.shape[1] for xk in X])

    # Calculate dimensionality for returned arrays
    n_tasks = len(X)
    n_features = max([xk.shape[1] for xk in X])

    # Build empty arrays
    cov_C = np.zeros((n_tasks, n_features))
    cov_D = np.zeros((n_tasks, n_features, n_features))

    # Populate arrays
    for task_id in range(n_tasks):
        cov_C[task_id] = np.dot(Y[task_id].transpose(), X[task_id])  # yTx
        cov_D[task_id] = np.dot(X[task_id].transpose(), X[task_id])  # xTx

    return cov_C, cov_D


def sum_squared_errors(X, Y, W, k):
    '''
    Get RSS for a particular task 'k'
    '''
    return(np.sum((Y[k].T-np.dot(X[k], W[:,k]))**2))


def ebic(X, Y, model_weights, n_tasks, n_samples, n_preds, gamma=1, min_rss=MIN_RSS):
    """
    Calculate EBIC for each task, and take the mean
    Extended Bayesian information criteria for model selection with large model spaces
    https://doi.org/10.1093/biomet/asn034
    :param X: list([N x K]) [T]
        List consisting of design matrixes for each task
    :param Y: list([N x 1]) [T]
        List consisting of response matrixes for each task
    :param model_weights: np.ndarray [K x T]
        Fit model coefficients for each task
    :param n_tasks: int
        Number of tasks T
    :param n_samples: list(int) [T]
        Number of samples for each task
    :param n_preds: int
        Number of predictors
    :param gamma: float
        Gamma parameter for extended BIC
    :param min_rss: float
        Floor value for RSS to prevent log(0)
    :return: float
        Mean ebic for all tasks
    """

    assert check.argument_type(X, list)
    assert check.argument_type(Y, list)
    assert check.argument_type(model_weights, np.ndarray)
    assert check.argument_type(n_tasks, int)
    assert check.argument_type(n_samples, list)
    assert check.argument_type(n_preds, int)
    assert check.argument_numeric(n_tasks, low=1)
    assert check.argument_numeric(n_preds, low=1)
    assert check.argument_type(gamma, (float, int))

    EBIC = []

    for task_id in range(n_tasks):
        # Get the number of samples for this task
        task_samples = n_samples[task_id]
        task_model = model_weights[:, task_id]

        # Find the number of non-zero predictors
        nonzero_pred = (task_model != 0).sum()

        # Calculate RSS for the task model
        rss = sum_squared_errors(X, Y, model_weights, task_id)

        # Calculate bayes information criterion using likelihood = RSS / n
        # Calculate the first component of BIC with a non-zero floor for RSS (so BIC is always finite)
        bic = task_samples * np.log(rss / task_samples) if rss > 0 else task_samples * np.log(min_rss / task_samples)

        # Calculate the second component of BIC
        bic_penalty = nonzero_pred * np.log(task_samples)

        # Calculate the extended component of eBIC
        # 2 * gamma * ln(number of non-zero predictor combinations of all predictors)
        bic_extension = 2 * gamma * np.log(comb(n_preds, nonzero_pred))

        # Combine all the components and put them in a list
        EBIC.append(bic + bic_penalty + bic_extension)

    return np.mean(EBIC)

def _final_weights(X, y, TFs, gene):
    """
    returns reduction on variance explained for each predictor
    (model without each predictor compared to full model)
    see: Greenfield et al., 2013. Robust data-driven incorporation of prior
    knowledge into the inference of dynamic regulatory networks.
    :param X: np.ndarray [N x k]
        A design matrix with N samples and k non-zero predictors
    :param y: np.ndarray [N x 1]
        A response matrix with N samples of a specific gene expression
    :param TFs: list() or np.ndarray or pd.Series
        A list of non-zero TFs (k) included in the model
    :param gene: str
        The gene modeled
    :return out_weights: pd.DataFrame [k x 4]
        An edge table (regulator -> target) with the model coefficient and the variance explained by that predictor for
        each non-zero predictor
    """

    assert check.argument_type(X, np.ndarray)
    assert check.argument_type(y, np.ndarray)
    assert check.argument_type(TFs, (list, np.ndarray, pd.Series))

    n_preds = len(TFs)

    # Linear fit using sklearn
    ols = LinearRegression().fit(X, y)

    # save weights and initialize rescaled weights vector
    weights = ols.coef_[0]
    resc_weights = np.zeros(n_preds)

    # variance of residuals (full model)
    var_full = np.var((y - ols.predict(X)) ** 2)

    # when there is only one predictor
    if n_preds == 1:
        resc_weights[0] = 1 - (var_full / np.var(y))
    # remove each at a time and calculate variance explained
    else:
        for j in range(len(TFs)):
            X_noj = X[:, np.setdiff1d(range(n_preds), j)]
            ols = LinearRegression().fit(X_noj, y)
            var_noj = np.var((y - ols.predict(X_noj)) ** 2)
            resc_weights[j] = 1 - (var_full / var_noj)

    # Format output into an edge table
    out_weights = pd.DataFrame([TFs, [gene] * len(TFs), weights, resc_weights]).transpose()
    out_weights.columns = ['regulator', 'target', 'weights', 'resc_weights']

    return out_weights


def format_prior(priors, gene, tasks, prior_weight):
    """
    Returns weighted priors for one gene
    :param priors: list(pd.DataFrame [G x K]) or pd.DataFrame [G x K]
        Either a list of prior data or a single data frame to use for all tasks
    :param gene: str
        The gene to select from the priors
    :param tasks: list(int)
        A list of task IDs
    :param prior_weight: float, int
        How much to weight the priors
    :return prior_out: np.ndarray [K x T]
        The weighted priors for a specific gene in each task
    """

    assert check.argument_type(priors, (list, pd.DataFrame), allow_none=True)
    assert check.argument_string(gene)
    assert check.argument_type(tasks, list)
    assert check.argument_numeric(prior_weight)

    if priors is None:
        return None

    priors_out = []

    # If the priors are a list, get the gene-specific prior from each task
    if isinstance(priors, list) and len(priors) > 1:
               
        for k in tasks:
            prior = priors[k].reindex([gene]).replace(np.nan, 0)
            priors_out.append(_weight_prior(prior.loc[gene, :], prior_weight))

    # Otherwise just use the same prior for each task
    else:

        priors = priors[0] if isinstance(priors, list) else priors
        
        prior = priors.reindex([gene]).replace(np.nan, 0)
        priors_out = [_weight_prior(prior.loc[gene, :], prior_weight)] * len(tasks)

    # Return a [K x T]
    priors_out = np.transpose(np.asarray(priors_out))
    return priors_out


def _weight_prior(prior, prior_weight):
    """
    Weight priors
    :param prior: pd.Series [K] or np.ndarray [K,]
        The prior information
    :param prior_weight: numeric
        How much to weight priors. If this is 1, do not weight prior terms differently
    :return prior: pd.Series [K] or np.ndarray [K,]
        Weighted priors
    """

    # Set non-zero priors to 1 and zeroed priors to 0
    prior = (prior != 0).astype(float)

    # Divide by the prior weight
    prior /= prior_weight

    # Set zeroed priors to 1
    prior[prior == 0] = 1.0

    # Reweight priors
    prior = prior / prior.sum() * len(prior)
    return prior


def _format_weights(df, col, targets, regs):
    """
    Reformat the edge table (target -> regulator) that's output by amusr into a pivoted table that the rest of the
    inferelator workflow can handle
    :param df: pd.DataFrame
        An edge table (regulator -> target) with columns containing model values
    :param col:
        Which column to pivot into values
    :param targets: list
        A list of target genes (the index of the output data)
    :param regs: list
        A list of regulators (the columns of the output data)
    :return out: pd.DataFrame [G x K]
        A [targets x regulators] dataframe pivoted from the edge dataframe
    """

    # Make sure that the value column is all numeric
    df[col] = pd.to_numeric(df[col])

    # Pivot an edge table into a matrix of values
    out = pd.pivot_table(df, index='target', columns='regulator', values=col, fill_value=0.)

    # Reindex to a [targets x regulators] dataframe and fill anything missing with 0s
    out = out.reindex(targets).reindex(regs, axis=1)
    out = out.fillna(value=0.)

    return out

def scale_list_of_arrays(X):
    """
    Scale a list of arrays so that each has mean 0 and unit variance
    
    :param X: list(np.ndarray) [T]
    :return X: list(np.ndarray) [T]
    """

    assert check.argument_type(X, list)

    return [StandardScaler().fit_transform(xk.astype(float)) for xk in X]

def scale_list_of_data(X):
    """
    Scale a list of data objects so that each has mean 0 and unit variance

    :param X: list(InferelatorData) [T]
    :return X: list(InferelatorData) [T]
    """

    assert check.argument_type(X, list)

    return [xk.zscore(ddof=0) for xk in X]

class AMUSRRegressionWorkflowMixin(base_regression._MultitaskRegressionWorkflowMixin):
    """
    Multi-Task AMuSR regression

    https://doi.org/10.1371/journal.pcbi.1006591
    """

    prior_weight = default.DEFAULT_prior_weight

    # Model hyperparameters
    lambda_Bs = None
    lambda_Ss = None
    heuristic_Cs = None

    def set_regression_parameters(self, prior_weight=None, lambda_Bs=None, lambda_Ss=None, heuristic_Cs=None):
        """
        Set regression parameters for AmUSR.

        :param prior_weight: Weight for edges that are present in the prior network.
            Non-prior edges have a weight of 1. Set this to 1 to weight prior and non-prior edges equally.
            Defaults to 1.
        :type prior_weight: numeric
        :param lambda_Bs: Lambda_B values to search during model selection.
            If not set, lambda_B will be chosen using the heuristic lambda_b = c * sqrt(d log p / n) from Castro 2019
            Defaults to not set. Must be provided if lambda_S is set.
        :type lambda_Bs: list(floats) or np.ndarray(floats)
        :param lambda_Ss: Lambda_S values to search during model selection.
            If not set, lambda_S will be chosen using the heuristic 0.5 < lambda_s/lambda_b < 1 from Castro 2019
            Defaults to not set.
        :type lambda_Ss: list(floats) or np.ndarray(floats)
        :param heuristic_Cs: c values to search during model selection.
            Values of c to calculate lambda_b = c * sqrt(d log p / n),
            Defaults to np.logspace(np.log10(0.01), np.log10(10), 20)[::-1].
            Does not have an effect if lambda_B is provided.
        :type heuristic_Cs: list(floats) or np.ndarray(floats)
        """

        self._set_with_warning("prior_weight", prior_weight)
        self._set_with_warning("lambda_Bs", lambda_Bs)
        self._set_with_warning("lambda_Ss", lambda_Ss)
        self._set_with_warning("heuristic_Cs", heuristic_Cs)


    def run_bootstrap(self, bootstrap_idx):
        x, y = [], []

        # Select the appropriate bootstrap from each task and stash the data into X and Y
        for k in range(self._n_tasks):
            x.append(self._task_design[k].get_bootstrap(self._task_bootstraps[k][bootstrap_idx]))
            y.append(self._task_response[k].get_bootstrap(self._task_bootstraps[k][bootstrap_idx]))

        MPControl.sync_processes(pref="amusr_pre")
        regress = AMuSR_regression(x, y, tfs=self._regulators, genes=self._targets, priors=self._task_priors,
                                   prior_weight=self.prior_weight, lambda_Bs=self.lambda_Bs, lambda_Ss=self.lambda_Ss, 
                                   Cs=self.heuristic_Cs)
        return regress.run()


def filter_genes_on_tasks(list_of_indexes, task_expression_filter):
    """
    Take a list of indexes and filter them based on the method specified in task_expression_filter to a single
    index
    :param list_of_indexes: list(pd.Index)
    :param task_expression_filter: str or int
    :return filtered_genes: pd.Index
    """

    filtered_genes = list_of_indexes[0]

    # If task_expression_filter is a number only keep genes in that number of tasks or higher
    if isinstance(task_expression_filter, int):
        filtered_genes = pd.concat(list(map(lambda x: x.to_series(), list_of_indexes))).value_counts()
        filtered_genes = filtered_genes[filtered_genes >= task_expression_filter].index
    # If task_expression_filter is "intersection" only keep genes in all tasks
    elif task_expression_filter == "intersection":
        for gene_idx in list_of_indexes:
            filtered_genes = filtered_genes.intersection(gene_idx)
    # If task_expression_filter is "union" keep genes that are in any task
    elif task_expression_filter == "union":
        for gene_idx in list_of_indexes:
            filtered_genes = filtered_genes.union(gene_idx)
    else:
        raise ValueError("{v} is not an allowed task_expression_filter value".format(v=task_expression_filter))

    return filtered_genes



import numpy as np
import pandas as pd

from inferelator import utils
from inferelator.utils import Validator as check
from inferelator.postprocessing.model_metrics import RankSummingMetric, MetricHandler
from inferelator.postprocessing.inferelator_results import InferelatorResults
from inferelator.postprocessing import (BETA_SIGN_COLUMN, MEDIAN_EXPLAIN_VAR_COLUMN, CONFIDENCE_COLUMN,
                                        BETA_THRESHOLD_COLUMN, TARGET_COLUMN, REGULATOR_COLUMN, PRIOR_COLUMN)

FILTER_METHODS = ("overlap", "keep_all_gold_standard")
DEFAULT_BOOTSTRAP_THRESHOLD = 0.5
DEFAULT_FILTER_METHOD = "overlap"
DEFAULT_METRIC = "precision-recall"


class ResultsProcessor(object):
    # Data
    betas = None
    rescaled_betas = None
    filter_method = DEFAULT_FILTER_METHOD

    # Processed Network
    network_data = None

    # Cutoffs
    threshold = DEFAULT_BOOTSTRAP_THRESHOLD

    # Flag to write results
    write_results = True

    # Model result object
    result_object = InferelatorResults

    # Model metric
    metric = None

    def __init__(self, betas, rescaled_betas, threshold=None, filter_method=None, metric=None):
        """
        :param betas: list(pd.DataFrame[G x K]) [B]
            A list of model weights per bootstrap
        :param rescaled_betas: list(pd.DataFrame[G x K]) [B]
            A list of the variance explained by each parameter per bootstrap
        :param threshold: float
            The proportion of bootstraps which an model weight must be non-zero for inclusion in the network output
        :param filter_method: str
            How to handle gold standard filtering ('overlap' filters to beta, 'keep_all_gold_standard' doesn't filter)
        :param metric: str / RankSummingMetric
            The scoring metric to use
        """

        self.validate_init_args(betas, rescaled_betas, threshold=threshold, filter_method=filter_method, metric=metric)

        self.betas = betas
        self.rescaled_betas = rescaled_betas
        self.filter_method = self.filter_method if filter_method is None else filter_method
        self.threshold = self.threshold if threshold is None else threshold

        metric = metric if metric is not None else DEFAULT_METRIC
        self.metric = MetricHandler.get_metric(metric)

    @staticmethod
    def validate_init_args(betas, rescaled_betas, threshold=None, filter_method=None, metric=None):
        assert check.argument_type(betas, list)
        assert check.argument_type(betas[0], pd.DataFrame)
        assert check.dataframes_align(betas)
        assert check.argument_type(rescaled_betas, list)
        assert check.argument_type(rescaled_betas[0], pd.DataFrame)
        assert check.dataframes_align(rescaled_betas)
        assert check.argument_enum(filter_method, FILTER_METHODS, allow_none=True)
        assert check.argument_numeric(threshold, 0, 1, allow_none=True)

    def summarize_network(self, output_dir, gold_standard, priors):
        """
        Take the betas and rescaled beta_errors, construct a network, and test it against the gold standard
        :param output_dir: str
            Path to write files into. Don't write anything if this is None.
        :param gold_standard: pd.DataFrame [G x K]
            Gold standard to test the network against
        :param priors: pd.DataFrame [G x K]
            Prior data
        :return result: InferelatorResult
            Returns an InferelatorResult
        """

        assert check.argument_path(output_dir, allow_none=True)
        assert check.argument_type(gold_standard, pd.DataFrame)
        assert check.argument_type(priors, pd.DataFrame)

        rs_calc = self.metric(self.rescaled_betas, gold_standard, filter_method=self.filter_method)
        beta_threshold, beta_sign, beta_nonzero = self.threshold_and_summarize(self.betas, self.threshold)
        resc_betas_mean, resc_betas_median = self.mean_and_median(self.rescaled_betas)
        extra_cols = {BETA_SIGN_COLUMN: beta_sign, MEDIAN_EXPLAIN_VAR_COLUMN: resc_betas_median}

        m_name, score = rs_calc.score()
        utils.Debug.vprint("Model {metric}:\t{score}".format(metric=m_name, score=score), level=0)

        # Process data into a network dataframe
        network_data = self.process_network(rs_calc, priors, beta_threshold=beta_threshold, extra_columns=extra_cols)

        # Create a InferelatorResult object and have it write output files
        result = self.result_object(network_data, beta_threshold, rs_calc.all_confidences, rs_calc,
                                    betas_sign=beta_sign, betas=self.betas)

        if self.write_results and output_dir is not None:
            result.write_result_files(output_dir)

        return result

    @staticmethod
    def process_network(metric, priors, confidence_threshold=0, beta_threshold=None, extra_columns=None):
        """
        Process rank-summed results into a network data frame
        :param metric: RankSummingMetric
            The rank-sum object with the math in it
        :param priors: pd.DataFrame [G x K]
            Prior data
        :param confidence_threshold: numeric
            The minimum confidence score needed to write a network edge
        :param beta_threshold: pd.DataFrame [G x K]
            The thresholded betas to include in the network. If None, include everything.
        :param extra_columns: dict(col_name: pd.DataFrame [G x K])
            Any additional data to include, keyed by column name and indexable with row and column names
        :return network_data: pd.DataFrame [(G*K) x 7+]
            Network edge dataframe

        """

        assert check.argument_type(metric, RankSummingMetric)
        assert check.argument_type(priors, pd.DataFrame, allow_none=True)
        assert check.argument_type(beta_threshold, pd.DataFrame, allow_none=True)
        assert check.argument_numeric(confidence_threshold, 0, 1)

        # Get the combined confidences and subset for confidence threshold
        network_data = metric.confidence_dataframe()
        network_data = network_data.loc[network_data[CONFIDENCE_COLUMN] > confidence_threshold, :]

        # If beta_threshold has been provided, melt and join it to the network data
        # Then discard anything which isn't meeting the threshold
        if beta_threshold is not None and False:
            beta_data = utils.melt_and_reindex_dataframe(beta_threshold, BETA_THRESHOLD_COLUMN)
            network_data = network_data.join(beta_data, on=[TARGET_COLUMN, REGULATOR_COLUMN])
            network_data = network_data.loc[network_data[BETA_THRESHOLD_COLUMN] == 1, :]
            del network_data[BETA_THRESHOLD_COLUMN]

        if priors is not None:
            prior_data = utils.melt_and_reindex_dataframe(priors, PRIOR_COLUMN)
            network_data = network_data.join(prior_data, on=[TARGET_COLUMN, REGULATOR_COLUMN])

        # Add any extra columns as needed
        if extra_columns is not None:
            for k in sorted(extra_columns.keys()):
                extra_data = utils.melt_and_reindex_dataframe(extra_columns[k], k)
                network_data = network_data.join(extra_data, on=[TARGET_COLUMN, REGULATOR_COLUMN])

        # Make sure all missing values are NaN
        network_data[pd.isnull(network_data)] = np.nan

        return network_data

    @staticmethod
    def threshold_and_summarize(betas, threshold):
        """
        Summarize a stack of betas
        Returns dataframes
        :param betas: list(pd.DataFrame)
            A list of dataframes that are aligned on both axes
        :param threshold: numeric
            The proportion of bootstraps an interaction must occur in to be valid
        :return betas_threshold: pd.DataFrame
        :return betas_sign: pd.DataFrame
        :return betas_non_zero: pd.DataFrame
        """
        betas_sign, betas_non_zero = ResultsProcessor.summarize(betas)
        betas_threshold = ResultsProcessor.passes_threshold(betas_non_zero, len(betas), threshold)
        return betas_threshold, betas_sign, betas_non_zero

    @staticmethod
    def summarize(betas):
        """
        Compute summary information about betas

        :param betas: list(pd.DataFrame) B x [M x N]
            A list of dataframes that are aligned on both axes
        :return betas_sign: pd.DataFrame [M x N]
            A dataframe with the summation of np.sign() for each bootstrap
        :return betas_non_zero: pd.DataFrame [M x N]
            A dataframe with a count of the number of non-zero betas for an interaction
        """

        assert check.dataframes_align(betas)

        betas_sign = pd.DataFrame(np.zeros(betas[0].shape), index=betas[0].index, columns=betas[0].columns)
        betas_non_zero = pd.DataFrame(np.zeros(betas[0].shape), index=betas[0].index, columns=betas[0].columns)
        for beta in betas:
            # Convert betas to -1,0,1 based on signing and then sum the results for each bootstrap
            betas_sign = betas_sign + np.sign(beta.values)
            # Tally all non-zeros for each bootstrap
            betas_non_zero = betas_non_zero + (beta != 0).astype(int)

        return betas_sign, betas_non_zero

    @staticmethod
    def passes_threshold(betas_non_zero, max_num, threshold):
        """

        :param betas_non_zero: pd.DataFrame [M x N]
            A dataframe of integer counts indicating how many times the original data was non-zero
        :param max_num: int
            The maximum number of possible counts (# bootstraps)
        :param threshold: float
            The proportion of integer counts over max possible in order to consider the interaction valid
        :return: pd.DataFrame [M x N]
            A bool dataframe where 1 corresponds to interactions that are in more than the threshold proportion of
            bootstraps
        """

        assert check.argument_type(betas_non_zero, pd.DataFrame)
        assert check.argument_integer(max_num)
        assert check.argument_numeric(threshold, low=0, high=1)

        return ((betas_non_zero / max_num) >= threshold).astype(int)

    @staticmethod
    def mean_and_median(stack):
        """
        Calculate the mean and median values of a list of dataframes
        Returns dataframes with the same dimensions as any one of the input stack
        :param stack: list(pd.DataFrame)
            List of dataframes which have the same size and dimensions
        :return mean_data: pd.DataFrame
            Mean values
        :return median_data:
            Median values
        """

        assert check.dataframes_align(stack)

        matrix_stack = [x.values for x in stack]
        mean_data = pd.DataFrame(np.mean(matrix_stack, axis=0), index=stack[0].index, columns=stack[0].columns)
        median_data = pd.DataFrame(np.median(matrix_stack, axis=0), index=stack[0].index, columns=stack[0].columns)
        return mean_data, median_data



import numpy as np
import pandas as pd
from abc import abstractmethod

from inferelator import utils

ISTS_COLUMN_NAME = 'isTs'
PREV_COLUMN_NAME = 'prevCol'
DELT_COLUMN_NAME = 'del.t'
COND_COLUMN_NAME = 'condName'

GROUP_COLUMN_NAME = "strain"
TIME_COLUMN_NAME = "time"

DEFAULT_STRICT_CHECKING_FOR_METADATA = False
DEFAULT_STRICT_CHECKING_FOR_DUPLICATES = True


class MetadataParser(object):

    cond_col = COND_COLUMN_NAME
    req_cols = tuple(COND_COLUMN_NAME)

    @classmethod
    @abstractmethod
    def process_groups(cls, meta_data):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def check_for_dupes(cls, exp_data, meta_data, steady_idx,
                        strict_checking_for_metadata=DEFAULT_STRICT_CHECKING_FOR_METADATA,
                        strict_checking_for_duplicates=DEFAULT_STRICT_CHECKING_FOR_DUPLICATES):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def validate_metadata(cls, exp_data, meta_data):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def create_default_meta_data(cls, metadata_index):
        raise NotImplementedError

    @classmethod
    def check_loaded_meta_data(cls, meta_data):
        if cls.cond_col in meta_data:
            meta_data.index = meta_data[cls.cond_col]
            meta_data.index.name = None
        return meta_data

    @classmethod
    def validate_metadata_columns(cls, meta_data):
        """
        Make sure metadata columns that are required for 'process_groups' exist
        :param meta_data: pd.DataFrame
        :return: True if the metadata is valid, False otherwise
        """
        cls.create_sample_name_column(meta_data)

        invalid_cols = pd.Index(cls.req_cols).difference(meta_data.columns)
        if len(invalid_cols) > 0:
            _msg = "Skipping metadata because parsing requires missing columns: {c}".format(c=invalid_cols)
            utils.Debug.vprint(_msg, level=0)
            return False
        else:
            return True

    @classmethod
    def create_sample_name_column(cls, meta_data):
        """
        Create a meta_data sample_name column from the index if necessary
        """
        if cls.cond_col not in meta_data:
            meta_data[cls.cond_col] = meta_data.index.astype(str)
        elif (meta_data[cls.cond_col] != meta_data.index.astype(str)).any():
            utils.Debug.vprint("Meta data sample name column and meta_data index are not equal", level=2)

    @staticmethod
    def fix_NAs(data_frame):
        """
        Replace the string NA with a np.nan
        :param data_frame: pd.DataFrame
        :return: pd.DataFrame
        """
        return data_frame.replace('NA', np.nan, regex=False)


class MetadataParserBranching(MetadataParser):
    """
    Metadata parser that handles prev_col & del_t metadata (which allows for Branching Time Courses)
    """

    ists_col = ISTS_COLUMN_NAME  # Column of booleans for "Is a time-series experiment"
    cond_col = COND_COLUMN_NAME  # Column of sample names (matching expression data column names)
    prev_col = PREV_COLUMN_NAME  # Column that identifies the previous timepoint sample name
    delt_col = DELT_COLUMN_NAME  # Column that identifies the delta time between this sample and the previous
    req_cols = (ISTS_COLUMN_NAME, COND_COLUMN_NAME, PREV_COLUMN_NAME, DELT_COLUMN_NAME)

    @classmethod
    def process_groups(cls, meta_data):
        """
        Parse the metadata to identify steady-state experiments and link timecourse experiments
        :param meta_data: pd.DataFrame
        :return steady_idx, ts_group: dict, dict
            steady_idx:  Dict keyed by condition, value is boolean (True if the condition is a steady-state experiment)
            ts_group: Dict keyed by condition. [(Previous_condition_name, Previous_delt),
                                                (Following_condition_name, Following_delt)]
        """
        time_series = meta_data[cls.ists_col].values.astype(bool)
        steadies = dict(zip(meta_data[cls.cond_col].astype(str), np.logical_not(time_series)))

        ts_data = meta_data[time_series].fillna(False)
        ts_dict = dict(zip(ts_data[cls.cond_col].astype(str).tolist(),
                           zip(ts_data[cls.prev_col].tolist(),
                               ts_data[cls.delt_col].tolist())))
        ts_group = {}
        for cond, (prev, delt) in ts_dict.items():
            if prev is False or delt is False:
                prev, delt = None, None
            try:
                op, od = ts_group[cond]
                ts_group[cond] = [(prev, delt), od]
            except KeyError:
                ts_group[cond] = [(prev, delt), (None, None)]
            if prev is not None:
                try:
                    op, od = ts_group[prev]
                    ts_group[prev] = [op, (cond, delt)]
                except KeyError:
                    ts_group[prev] = [(None, None), (cond, delt)]

        return steadies, ts_group

    @classmethod
    def check_for_dupes(cls, exp_data, meta_data, steady_idx,
                        strict_checking_for_metadata=DEFAULT_STRICT_CHECKING_FOR_METADATA,
                        strict_checking_for_duplicates=DEFAULT_STRICT_CHECKING_FOR_DUPLICATES):
        """
        Sanity check the metadata and experimental data to ensure that they can be linked. Try to guess as much as
        possible unless strict checking is on
        """
        # Check to make sure that the conditions in the expression data are matched with conditions in the metadata
        exp_conds = exp_data.columns.astype(str)
        meta_conds = meta_data[cls.cond_col].astype(str)

        # Check to find out if there are any conditions which don't have associated metadata
        # If there are, raise an exception if strict_checking_for_metadata is True
        # Otherwise just assume they're steady-state data and move on
        in_exp_not_meta = exp_conds.difference(meta_conds).tolist()
        if len(in_exp_not_meta) != 0:
            utils.Debug.vprint("{n} conditions cannot be properly matched to metadata".format(n=len(in_exp_not_meta)),
                               level=1)
            utils.Debug.vprint(" ".join(in_exp_not_meta), level=2)
            if strict_checking_for_metadata:
                raise ConditionDoesNotExistError("Conditions exist without associated metadata")
            else:
                for condition in in_exp_not_meta:
                    steady_idx[condition] = True

        # Check to find out if the conditions in the expression data are all unique
        # It's not a problem if they're not, we just assume the metadata applies to all conditions with the same name
        duplicate_in_exp = exp_conds.duplicated()
        n_dup_in_exp = np.sum(duplicate_in_exp)
        if n_dup_in_exp > 0:
            c_dup = exp_conds[duplicate_in_exp].tolist()
            utils.Debug.vprint("The expression data has {n} non-unique condition indexes".format(n=n_dup_in_exp),
                               level=1)
            utils.Debug.vprint(" ".join(c_dup), level=2)

        # Check to find out if the conditions in the meta data are all unique
        # If there are repeats, check and see if the associated information is identical (if it is, just ignore it)
        # If there are repeated conditions with different characteristics, the outcome may be unexpected
        # (The parser will just overwrite the first ones it comes to with the characteristics of the last one)
        duplicate_in_meta = meta_conds.duplicated()
        n_dup_in_meta = np.sum(duplicate_in_meta)
        if n_dup_in_meta > 0:
            meta_dup = meta_conds[duplicate_in_meta].tolist()
            if np.sum(np.logical_xor(meta_data.duplicated(), duplicate_in_meta)) > 0:
                utils.Debug.vprint("The metadata has non-unique conditions with different characteristics:", level=0)
                utils.Debug.vprint(" ".join(meta_dup), level=0)
                if strict_checking_for_duplicates:
                    raise MultipleConditionsError("Identical conditions have non-identical characteristics")
            else:
                utils.Debug.vprint("The metadata contains {n} duplicate rows".format(n=n_dup_in_meta), level=1)
                utils.Debug.vprint(" ".join(meta_dup), level=2)

        return steady_idx

    @classmethod
    def validate_metadata(cls, exp_data, meta_data):
        """
        Make sure that meta_data and expression_data are compatible
        """

        # Check and make sure that there's a name column and create it if needed
        cls.create_sample_name_column(meta_data)

        # Check the alignment of the expression data and the meta_data
        sample_names_expr = exp_data.columns.astype(str)
        sample_names_meta = meta_data[cls.cond_col].astype(str)
        align_count = len(sample_names_expr.intersection(sample_names_meta))

        if align_count == 0:
            raise ConditionDoesNotExistError("Unable to align metadata to expression data")
        elif align_count < min(exp_data.shape[1], meta_data.shape[0]):
            utils.Debug.vprint("Metadata ({me}) and expression data ({ex}) alignment off".format(me=meta_data.shape,
                                                                                                 ex=exp_data.shape),
                               level=0)

    @classmethod
    def create_sample_name_column(cls, meta_data):
        """
        Create a meta_data sample_name column from the index if necessary
        """
        if cls.cond_col not in meta_data:
            meta_data[cls.cond_col] = meta_data.index.astype(str)
        elif (meta_data[cls.cond_col] != meta_data.index.astype(str)).any():
            utils.Debug.vprint("Meta data sample name column and meta_data index are not equal", level=2)

    @classmethod
    def create_default_meta_data(cls, metadata_index):
        """
        Create a meta_data dataframe from basic defaults
        """

        # Create an empty dataframe with index equal to sample names from expression data
        meta_data = pd.DataFrame(index=metadata_index.astype(str))

        # Create a name column
        cls.create_sample_name_column(meta_data)

        # Assign default values to information columns
        meta_data[cls.ists_col] = True
        meta_data[cls.prev_col] = "NA"
        meta_data[cls.delt_col] = "NA"
        return meta_data


class MetadataParserNonbranching(MetadataParserBranching):
    group_col = GROUP_COLUMN_NAME
    time_col = TIME_COLUMN_NAME
    cond_col = COND_COLUMN_NAME
    req_cols = (GROUP_COLUMN_NAME, TIME_COLUMN_NAME, COND_COLUMN_NAME)

    @classmethod
    def process_groups(cls, meta_data):
        ts_dict = dict(zip(meta_data[cls.cond_col].astype(str).tolist(),
                           zip(meta_data[cls.group_col].tolist(),
                               meta_data[cls.time_col].tolist())))

        times_per_group = {k: [] for k in meta_data[cls.group_col].unique()}
        group_time_cond = {k: {} for k in meta_data[cls.group_col].unique()}
        ts_group = {cond: [(None, None), (None, None)] for cond in ts_dict.keys()}
        steady_idx = {cond: False for cond in ts_dict.keys()}

        # Walk through metadata to find the experiment times for each strain
        # And the sample IDs associated with each strain/time combo
        for cond, (group, time) in ts_dict.items():
            times_per_group[group].append(time)
            group_time_cond[group][time] = cond
            if np.isnan(time):
                steady_idx[cond] = True

        # Process into a dict, keyed by sample ID, of [(previous sample, del.t), (next sample, del.t)]
        for cond, (group, time) in ts_dict.items():
            group_times = times_per_group[group]

            if (min(group_times) == time and max(group_times) == time) or steady_idx[cond]:
                steady_idx[cond] = True
                ts_group.pop(cond)
                continue

            if min(group_times) == time:
                prev = (None, None)
            else:
                prev_time = max([i for (i, v) in zip(group_times, map(lambda x: x < time, group_times)) if v])
                prev_cond = group_time_cond[group][prev_time]
                prev = (prev_cond, time - prev_time)

            if max(group_times) == time:
                nex = (None, None)
            else:
                nex_time = min([i for (i, v) in zip(group_times, map(lambda x: x > time, group_times)) if v])
                nex_cond = group_time_cond[group][nex_time]
                nex = (nex_cond, nex_time - time)

            ts_group[cond] = [prev, nex]

        return steady_idx, ts_group

    @classmethod
    def create_default_meta_data(cls, metadata_index):
        """
        Create a meta_data dataframe from basic defaults
        """
        meta_data = pd.DataFrame(index=metadata_index.astype(str))
        meta_data[cls.cond_col] = metadata_index.astype(str)
        meta_data[cls.group_col] = list(range(len(metadata_index)))
        meta_data[cls.time_col] = 0
        return meta_data


class MetadataHandler(object):
    """
    This keeps track of how to process metadata
    """

    @classmethod
    def get_handler(cls, handler_ref):
        """
        This wrappers a metadata reference so that strings can be used instead of python imports
        Will either return a metadata handling class or will raise an error
        :param handler_ref: str / MetadataParser
            String or subclass of MetadataParser
        :return: MetadataParser
            The metadata parser that corresponds to the string, or the MetadataParser object will be passed through
        """
        if utils.is_string(handler_ref):
            if handler_ref == "branching":
                return MetadataParserBranching
            elif handler_ref == "nonbranching":
                return MetadataParserNonbranching
            else:
                raise ValueError("Parser {parser_str} unknown".format(parser_str=handler_ref))
        elif issubclass(handler_ref, MetadataParser):
            return handler_ref
        else:
            raise ValueError("Handler must be a string or a MetadataParser class")


class ConditionDoesNotExistError(IndexError):
    pass


class MultipleConditionsError(ValueError):
    pass
